# Import packages
import joblib
import numpy as np

def vectorField(location, isNode = False, nodeNumber = -1):
    """
    calculates vector fields

    Parameters
    ----------
    location : int or list
        location of point where vector field is to be calculated
    isNode : boolean
        to mention if selected point is node or not
    nodeNumber : int
        number of the node if node is selected

    Returns
    -------
    field : list
        list of x and y components of vector fields
    """

    if isinstance(location, int):
        domains = joblib.load("Data/initialDomains.sav")
        if domains[location].shape == 'point':
            nodeNumber = domains[location].boundaryList[0]
            return(vectorField(np.array(domains[location].position), True, nodeNumber))
        elif domains[location-1].shape == 'line' and domains[location].shape == 'line' and domains[location+1].shape == 'line':
            print("Field along a line is in progress")
            mesh = joblib.load("Data/mesh.sav")
            solver = joblib.load("Data/solver.sav")
            x = mesh.x
            y = mesh.y
            tri = mesh.tri
            u = solver.u

            l1 = domains[location-1].boundaryList[0]
            l2 = domains[location].boundaryList[0]
            l3 = domains[location+1].boundaryList[0]

            field = []
            for i in range(1,len(l2)-1):
                Exp = (l2[i]-l2[i+1])/(np.sqrt((x[l2[i]] - x[l2[i+1]])**2+(y[l2[i]] - y[l2[i+1]])**2))
                Exn = (l2[i-1]-l2[i])/(np.sqrt((x[l2[i-1]] - x[l2[i]])**2+(y[l2[i-1]] - y[l2[i]])**2))
                Ex = (Exp+Exn)/2
                Eyp = (l2[i]-l1[i])/(np.sqrt((x[l2[i]] - x[l1[i]])**2+(y[l2[i]] - y[l1[i]])**2))
                Eyn = (l3[i]-l2[i])/(np.sqrt((x[l3[i]] - x[l2[i]])**2+(y[l3[i]] - y[l2[i]])**2))
                Ey = (Eyp+Eyn)/2
                field.append([Ex, Ey])

            return field

        else:
            print("Field along the line cannot be obtained for the selected domain", location)
    elif isinstance(location, np.ndarray):
        domains = joblib.load("Data/initialDomains.sav")
        mesh = joblib.load("Data/mesh.sav")
        solver = joblib.load("Data/solver.sav")
        x = mesh.x
        y = mesh.y
        tri = mesh.tri
        u = solver.u

        if isNode:
            field = [0, 0]
            # node = domains[domainNumber].boundaryList[0]
            eleAttNames = mesh.eleAttNames.copy()
            neiElements = getattr(mesh, eleAttNames[0])[nodeNumber]
            elementfields = []
            for element in neiElements:
                A = [[0, 0, 0], [0, 0, 0], [0, 0, 0]]
                for k in range(3):
                    A[k] = [1, x[tri[element][k]], y[tri[element][k]]]
                b = [[u[tri[element][0]][0]], [u[tri[element][1]][0]], [u[tri[element][2]][0]]]
                X = np.linalg.lstsq(A, b)[0]
                elementfield = [-X[1][0], -X[2][0]]
                elementfields.append(elementfield)
            elementfields = np.array(elementfields)
            field = sum(elementfields)/len(elementfields)
        else:
            field = []
            noOfRows = location.shape[0]
            for i in range(noOfRows):
                field.append([0, 0])

            for j in range(noOfRows):
                isInsideGlobalBoundary = False
                for i in range(len(tri)):
                    if isInside(x[tri[i][0]], y[tri[i][0]], x[tri[i][1]], y[tri[i][1]], x[tri[i][2]], y[tri[i][2]], location[j][0], location[j][1]):
                        theElement = i
                        isInsideGlobalBoundary = True
                        break

                if isInsideGlobalBoundary:
                    A = [[0, 0, 0], [0, 0, 0], [0, 0, 0]]
                    for k in range(3):
                        A[k] = [1, x[tri[theElement][k]], y[tri[theElement][k]]]
                    b = [[u[tri[theElement][0]][0]], [u[tri[theElement][1]][0]], [u[tri[theElement][2]][0]]]
                    X = np.linalg.lstsq(A, b)[0]
                    field[j] = [-X[1][0], -X[2][0]]
                else:
                    field[j] = [[0, 0]]
                    print("Position", location[j], "is out of Global Boundary")

        if solver.name == "magnetostatic":
            print("Turn field 90 degrees")
        return field

def area(x1, y1, x2, y2, x3, y3):
    """
    calculates area

    Parameters
    ----------
    x1 : float
        x position of vertix 1
    y1 : float
        y position of vertix 1
    x2 : float
        x position of vertix 2
    y2 : float
        y position of vertix 2
    x3 : float
        x position of vertix 3
    y3 : float
        y position of vertix 3
    
    Returns
    -------
    area : float
        area of triangle
    """

    return abs((x1 * (y2 - y3) + x2 * (y3 - y1)+ x3 * (y1 - y2))/ 2.0)

def isInside(x1, y1, x2, y2, x3, y3, x, y):
    """
    calculates if a point is inside a triangle or not

    Parameters
    ----------
    x1 : float
        x position of vertix 1
    y1 : float
        y position of vertix 1
    x2 : float
        x position of vertix 2
    y2 : float
        y position of vertix 2
    x3 : float
        x position of vertix 3
    y3 : float
        y position of vertix 3
    x : float
        x position of the point
    y : float
        y position of the point
    
    Returns
    -------
    isInside : boolean
        returns true if point is inside
    """

    A = area(x1, y1, x2, y2, x3, y3)
    A1 = area(x, y, x2, y2, x3, y3)
    A2 = area(x1, y1, x, y, x3, y3)
    A3 = area(x1, y1, x2, y2, x, y)
    AT = A1 + A2 + A3
    if(AT < A*1.00001): # To avoid floating point error
        return True
    else:
        return False

def force(domainNumber):
    """
    calculates force on a domain

    Parameters
    ----------
    domainNumber: int
        number of domain for which force is to be calculated

    Returns
    -------
    Force : list
        list of x and y components of vector Force
    """

    print("Get.force is under progress")
    domains = joblib.load("Data/initialDomains.sav")
    mesh = joblib.load("Data/mesh.sav")
    solver = joblib.load("Data/solver.sav")
    x = mesh.x
    y = mesh.y
    nodeSource = mesh.nodeSource
    nodes = np.concatenate((domains[domainNumber].boundaryList, domains[domainNumber].freeList)).astype(int)
    nodeForce = []
    for node in nodes:
        nodeVectorField = vectorField(np.array([x[node], y[node]]), True, node)
        if solver.name == "electrostatic":
            nodeForce.append(nodeVectorField*nodeSource[node])
        elif solver.name == "magnetostatic":
            print("turn magnetic force 90 degrees")
            nodeForce.append(nodeVectorField*nodeSource[node])
    Force = sum(nodeForce)
    return Force




import numpy as np

from . import Domain

class Line(Domain.Domain):
    def __init__(self, width, height, position, relativeMeshDensity, angle, isChild, isParent, childDomains, parentDomains, non):
        """
        Constructs all the necessary attributes for the geometry object.

        Parameters
        ----------
        shape: str
            the name of the shape of geometry
        """
        
        Domain.Domain.__init__(self, width, height, position, relativeMeshDensity, angle, isChild, isParent, childDomains, parentDomains, non)
        self.shape = "line"

    def show(self, plt, lW):
        """
        adds the current geometry to plot

        Parameters
        ----------
        plt : object
            matplotlib object
        lw : float
            width of line stroke

        Returns
        -------
        None
        """

        # Generating boundary nodes for plotting
        w = self.width
        v0 = np.array([[0,0],[w,0]])
        v0 = np.transpose(v0)
        # Rotate the boundary points
        thr = self.angle
        R = np.array([[np.cos(np.radians(thr)), -np.sin(np.radians(thr))],
        [np.sin(np.radians(thr)), np.cos(np.radians(thr))]])
        vr = np.dot(R,v0)
        # Reposition the boundary points
        position = np.array(self.position)
        v = position +  np.transpose(vr)
        # Plot the boundary points
        plt.plot(v[[0,1],0], v[[0,1],1], 'k', lw = lW)

    def calculateArea(self):
        """
        calculates area.

        Parameters
        ----------
        None

        Returns
        -------
        None
        """

        self.area = 0

    def calculateMaxFreeNodes(self, domains):
        """
        calculates maximum number of free nodes allowed

        Parameters
        ----------
        domains : list
            list of all geometry objects

        Returns
        -------
        None
        """

        self.maxFreeNodes = 0

    def genVertices(self, plt, lW, markerShape):
        """
        gets vertices of the geometry.

        Parameters
        ----------
        plt : object
            matplotlib object
        lw : float
            width of line stroke
        markerShape : char
            shape of marker

        Returns
        -------
        None
        """

        # Generating vertices
        w = self.width
        v0 = np.array([[0,0], [w,0]])
        v0 = np.transpose(v0)
        # Rotate the boundary points
        thr = self.angle
        R = np.array([[np.cos(np.radians(thr)), -np.sin(np.radians(thr))],
        [np.sin(np.radians(thr)), np.cos(np.radians(thr))]])
        vr = np.dot(R,v0)
        # Reposition the boundary points
        position = np.array(self.position)
        v = np.transpose(position +  np.transpose(vr))
        self.verticesX = v[0]
        self.verticesY = v[1]
        # Plot the vertices
        plt.scatter(self.verticesX, self.verticesY, c=None, marker=markerShape, s = lW*1000)

    def genBoundaryNodes(self, plt, lW, markerShape):
        """
        gets boundary nodes.

        Parameters
        ----------
        plt : object
            matplotlib object
        lw : float
            width of line stroke
        markerShape : char
            shape of marker

        Returns
        -------
        None
        """

        self.boundaryNodesX = np.linspace(self.verticesX[0],self.verticesX[1],self.widthNodes)
        self.boundaryNodesY = np.linspace(self.verticesY[0],self.verticesY[1],self.widthNodes)
        plt.scatter(self.boundaryNodesX, self.boundaryNodesY, c=None, marker=markerShape, s = lW*1000)

    def genRandFreeNodes(self, initialDomains, plt, lW, markerShape):
        """
        gets random free nodes inside a domain.

        Parameters
        ----------
        initialDomains : list
            list of all geometry objects
        plt : object
            matplotlib object
        lw : float
            width of line stroke
        markerShape : char
            shape of marker

        Returns
        -------
        None
        """

        self.freeNodesX = np.array([])
        self.freeNodesY = np.array([])
        self.allNodesX = self.boundaryNodesX
        self.allNodesY = self.boundaryNodesY

    def genTriFreeNodes(self, initialDomains, plt, lW, markerShape):
        """
        gets uniformly distributed triangular free nodes inside a domain.

        Parameters
        ----------
        initialDomains : list
            list of all geometry objects
        plt : object
            matplotlib object
        lw : float
            width of line stroke
        markerShape : char
            shape of marker

        Returns
        -------
        None
        """

        self.freeNodesX = np.array([])
        self.freeNodesY = np.array([])
        self.allNodesX = self.boundaryNodesX
        self.allNodesY = self.boundaryNodesY


from . import Solver
import numpy as np

class Magnetostatic(Solver.Solver):
    """
    A subclass to represent a specific Solver.
    
    ...

    Attributes
    ----------
    name : str
        name of the solver
    physicalConstant : float
        physical constant associated with the solver
    proportionalityConstant : float
        proportionality constant associated with the solver
    
    Methods
    -------
    getApproximateMatrices(mesh, initialDomains):
        gets approximate matrices
    getLinNumericalMatrix(mesh):
        gets linear numerical matrices
    getNumericalMatrix(mesh):
        gets non linear numerical matrices
    getApproximateBoundaryMatrices(mesh, initialDomains):
        gets approximate matrices on boundary
    getIterBoundaryMatrices(mesh, initialDomains, bt):
        gets approximate matrices on boundary for iterative update
    getMiniLinNumericalMatrix(minix, miniy, miniTri):
        gets linear numerical matrices
    findVectorFields(mesh, solver):
        finds gradient vector fields
    """

    def __init__(self):
        """
        A subclass to represent a specific Solver.
    
        Parameters
        ----------
        name : str
            name of the solver
        physicalConstant : float
            physical constant associated with the solver
        proportionalityConstant : float
            proportionality constant associated with the solver
        """

        Solver.Solver.__init__(self)
        self.name = "magnetostatic"
        self.physicalConstant = 12.5663706144*10**-7
        self.proportionalityConstant = self.physicalConstant/(2*np.pi)

    def getApproximateMatrices(self, mesh, initialDomains):
        """
        gets approximate matrices.

        Parameters
        ----------
        mesh : object
            the mesh object
        initialDomains : list
            list of geometries

        Returns
        -------
        None
        """

        x = mesh.rx
        y = mesh.ry
        tri = mesh.tri
        nn = len(x)
        ne = len(tri[:,0])
        pc = self.proportionalityConstant
        A = np.zeros((nn,nn))
        b = np.zeros((nn,1))

        for i in range(ne):

            print("Percentage of elements covered", i/ne*100, end='\r')

            # 0-1 edge
            k = 0
            l = 1
            Ex = 0
            Ey = 0

            for domain in initialDomains:
                if domain.source != 0:
                    if domain.shape == "point" or domain.shape == "line":
                        nodes = domain.boundaryList
                        for node in nodes:
                            px = x[node]
                            py = y[node]
                            Rx = (x[tri[i,k]] + x[tri[i,l]])/2 - px
                            Ry = (y[tri[i,k]] + y[tri[i,l]])/2 - py
                            Ex = Ex + mesh.nodeSource[node]*pc*Rx/(np.sqrt(Rx**2 + Ry**2))**2
                            Ey = Ey + mesh.nodeSource[node]*pc*Ry/(np.sqrt(Rx**2 + Ry**2))**2

                    else:
                        nodes = np.concatenate((domain.boundaryList, domain.freeList)).astype(int)
                        for node in nodes:
                            px = x[node]
                            py = y[node]
                            Rx = (x[tri[i,k]] + x[tri[i,l]])/2 - px
                            Ry = (y[tri[i,k]] + y[tri[i,l]])/2 - py
                            Ex = Ex + mesh.nodeSource[node]*pc*Rx/(np.sqrt(Rx**2 + Ry**2))**2
                            Ey = Ey + mesh.nodeSource[node]*pc*Ry/(np.sqrt(Rx**2 + Ry**2))**2

            rx = x[tri[i,l]] - x[tri[i,k]]
            ry = y[tri[i,l]] - y[tri[i,k]]
            Er = np.dot([Ex[0], Ey[0]], [rx, ry]/np.sqrt(rx**2+ry**2))
            A[tri[i,k], tri[i,k]] = A[tri[i,k], tri[i,k]] +1
            A[tri[i,k], tri[i,l]] = -1
            b[tri[i,k]]           = b[tri[i,k]] + Er*np.sqrt(rx**2+ry**2)

            # 1-2 edge
            k = 1
            l = 2
            Ex = 0
            Ey = 0

            for domain in initialDomains:
                if domain.source != 0:
                    if domain.shape == "point" or domain.shape == "line":
                        nodes = domain.boundaryList
                        for node in nodes:
                            px = x[node]
                            py = y[node]
                            Rx = (x[tri[i,k]] + x[tri[i,l]])/2 - px
                            Ry = (y[tri[i,k]] + y[tri[i,l]])/2 - py
                            Ex = Ex + mesh.nodeSource[node]*pc*Rx/(np.sqrt(Rx**2 + Ry**2))**2
                            Ey = Ey + mesh.nodeSource[node]*pc*Ry/(np.sqrt(Rx**2 + Ry**2))**2

                    else:
                        nodes = np.concatenate((domain.boundaryList, domain.freeList)).astype(int)
                        for node in nodes:
                            px = x[node]
                            py = y[node]
                            Rx = (x[tri[i,k]] + x[tri[i,l]])/2 - px
                            Ry = (y[tri[i,k]] + y[tri[i,l]])/2 - py
                            Ex = Ex + mesh.nodeSource[node]*pc*Rx/(np.sqrt(Rx**2 + Ry**2))**2
                            Ey = Ey + mesh.nodeSource[node]*pc*Ry/(np.sqrt(Rx**2 + Ry**2))**2

            rx = x[tri[i,l]] - x[tri[i,k]]
            ry = y[tri[i,l]] - y[tri[i,k]]
            Er = np.dot([Ex[0], Ey[0]], [rx, ry]/np.sqrt(rx**2+ry**2))
            A[tri[i,k], tri[i,k]] = A[tri[i,k], tri[i,k]] +1
            A[tri[i,k], tri[i,l]] = -1
            b[tri[i,k]]           = b[tri[i,k]] + Er*np.sqrt(rx**2+ry**2)

            # 2-0 edge
            k = 2
            l = 0
            Ex = 0
            Ey = 0

            for domain in initialDomains:
                if domain.source != 0:
                    if domain.shape == "point" or domain.shape == "line":
                        nodes = domain.boundaryList
                        for node in nodes:
                            px = x[node]
                            py = y[node]
                            Rx = (x[tri[i,k]] + x[tri[i,l]])/2 - px
                            Ry = (y[tri[i,k]] + y[tri[i,l]])/2 - py
                            Ex = Ex + mesh.nodeSource[node]*pc*Rx/(np.sqrt(Rx**2 + Ry**2))**2
                            Ey = Ey + mesh.nodeSource[node]*pc*Ry/(np.sqrt(Rx**2 + Ry**2))**2

                    else:
                        nodes = np.concatenate((domain.boundaryList, domain.freeList)).astype(int)
                        for node in nodes:
                            px = x[node]
                            py = y[node]
                            Rx = (x[tri[i,k]] + x[tri[i,l]])/2 - px
                            Ry = (y[tri[i,k]] + y[tri[i,l]])/2 - py
                            Ex = Ex + mesh.nodeSource[node]*pc*Rx/(np.sqrt(Rx**2 + Ry**2))**2
                            Ey = Ey + mesh.nodeSource[node]*pc*Ry/(np.sqrt(Rx**2 + Ry**2))**2

            rx = x[tri[i,l]] - x[tri[i,k]]
            ry = y[tri[i,l]] - y[tri[i,k]]
            Er = np.dot([Ex[0], Ey[0]], [rx, ry]/np.sqrt(rx**2+ry**2))
            A[tri[i,k], tri[i,k]] = A[tri[i,k], tri[i,k]] +1
            A[tri[i,k], tri[i,l]] = -1
            b[tri[i,k]]           = b[tri[i,k]] + Er*np.sqrt(rx**2+ry**2)

        print()
        self.Aa = A
        self.ba = b

    def getLinNumericalMatrix(self, mesh):
        """
        gets linear numerical matrices

        Parameters
        ----------
        mesh : object
            the mesh object

        Returns
        -------
        None
        """

        x = mesh.rx
        y = mesh.ry
        tri = mesh.tri
        nn = len(x)
        A = np.zeros((nn,nn))

        count = 0
        ne = len(tri[:,0])
        for i in range(ne):
            print("percentage of linear elements covered", count/ne*100, end='\r')
            count = count +1
            ex = np.array([x[tri[i,0]], x[tri[i,1]], x[tri[i,2]]])
            ey = np.array([y[tri[i,0]], y[tri[i,1]], y[tri[i,2]]])
            p = np.array([ey[1]-ey[2], ey[2]-ey[0], ey[0]-ey[1]])
            q = np.array([ex[2]-ex[1], ex[0]-ex[2], ex[1]-ex[0]])
            ea = 0.5*abs(p[1]*q[2]-q[1]*p[2])
            P = np.zeros([len(p),len(p)])
            Q = P.copy()
            for j in range(len(p)):
                P[j] = p*p[j]
                Q[j] = q*q[j]

            a = (P+Q)/(4*ea)
            for k in range(3):
                for j in range(3):
                    A[tri[i,k],tri[i,j]] = A[tri[i,k],tri[i,j]].copy() + a[k,j]

        print()
        self.AnLin = A.copy()

    def getNumericalMatrix(self, mesh):
        """
        gets non linear numerical matrices
        Parameters
        ----------
        mesh : object
            the mesh object

        Returns
        -------
        None
        """

        x = mesh.rx
        y = mesh.ry
        tri = mesh.tri
        nn = len(x)
        A = np.zeros((nn,nn))
        ur = mesh.elementMaterial

        count = 0
        ne = len(tri[:,0])
        for i in range(ne):
            print("percentage of non-linear elements covered", count/ne*100, end='\r')
            count = count +1
            ex = np.array([x[tri[i,0]], x[tri[i,1]], x[tri[i,2]]])
            ey = np.array([y[tri[i,0]], y[tri[i,1]], y[tri[i,2]]])
            p = np.array([ey[1]-ey[2], ey[2]-ey[0], ey[0]-ey[1]])
            q = np.array([ex[2]-ex[1], ex[0]-ex[2], ex[1]-ex[0]])
            ea = 0.5*abs(p[1]*q[2]-q[1]*p[2])
            P = np.zeros([len(p),len(p)])
            Q = P.copy()
            for j in range(len(p)):
                P[j] = p*p[j]
                Q[j] = q*q[j]

            a = (P+Q)/(4*ea*ur[i])
            for k in range(3):
                for j in range(3):
                    A[tri[i,k],tri[i,j]] = A[tri[i,k],tri[i,j]].copy() + a[k,j]

        print()
        self.An = A.copy()

    def getApproximateBoundaryMatrices(self, mesh, initialDomains):
        """
        gets approximate matrices on boundary

        Parameters
        ----------
        mesh : object
            the mesh object
        initialDomains : list
            list of geometries

        Returns
        -------
        None
        """

        x = mesh.rx
        y = mesh.ry
        tri = mesh.tri
        pc = self.proportionalityConstant

        # Boundary nodes
        gc = initialDomains[0].childDomains
        lbn = initialDomains[0].boundaryList
        lbn = np.append([lbn], [initialDomains[0].freeList])
        lbn = np.append([lbn], [initialDomains[gc[0]].boundaryList])
        lbn = lbn.astype(int)
        bx = x[lbn].copy()
        by = y[lbn].copy()

        # Boundary elements
        belements = initialDomains[0].elementList
        btri = tri[belements].copy()
        bne = len(btri[:,0])
        bnn = len(bx)

        # Mapping btri and tri

        for j in range(bne):
            for k in range(3):
                gn = btri[j,k]

                for l in range(bnn):
                    if gn == lbn[l]:
                        btri[j,k] = l
                        break
        
        # Initializing matrices
        A = np.zeros([bnn,bnn])
        b = np.zeros([bnn,1])

        # For every boundary element
        for i in range(bne):

            print("Percentage of boundary elements covered", i/bne*100, end='\r')

            # 0-1 edge
            k = 0
            l = 1
            Ex = 0
            Ey = 0

            for domain in initialDomains:
                if domain.source != 0:
                    if domain.shape == "point" or domain.shape == "line":
                        nodes = domain.boundaryList
                        for node in nodes:
                            px = x[node]
                            py = y[node]
                            Rx = (bx[btri[i,k]] + bx[btri[i,l]])/2 - px
                            Ry = (by[btri[i,k]] + by[btri[i,l]])/2 - py
                            Ex = Ex + mesh.nodeSource[node]*pc*Rx/(np.sqrt(Rx**2 + Ry**2))**2
                            Ey = Ey + mesh.nodeSource[node]*pc*Ry/(np.sqrt(Rx**2 + Ry**2))**2

                    else:
                        nodes = np.concatenate((domain.boundaryList, domain.freeList)).astype(int)
                        for node in nodes:
                            px = x[node]
                            py = y[node]
                            Rx = (bx[btri[i,k]] + bx[btri[i,l]])/2 - px
                            Ry = (by[btri[i,k]] + by[btri[i,l]])/2 - py
                            Ex = Ex + mesh.nodeSource[node]*pc*Rx/(np.sqrt(Rx**2 + Ry**2))**2
                            Ey = Ey + mesh.nodeSource[node]*pc*Ry/(np.sqrt(Rx**2 + Ry**2))**2

            rx = bx[btri[i,l]] - bx[btri[i,k]]
            ry = by[btri[i,l]] - by[btri[i,k]]
            Er = np.dot([Ex[0], Ey[0]], [rx, ry]/np.sqrt(rx**2+ry**2))
            A[btri[i,k], btri[i,k]] = A[btri[i,k], btri[i,k]] +1
            A[btri[i,k], btri[i,l]] = -1
            b[btri[i,k]]           = b[btri[i,k]] + Er*np.sqrt(rx**2+ry**2)

            # 1-2 edge
            k = 1
            l = 2
            Ex = 0
            Ey = 0

            for domain in initialDomains:
                if domain.source != 0:
                    if domain.shape == "point" or domain.shape == "line":
                        nodes = domain.boundaryList
                        for node in nodes:
                            px = x[node]
                            py = y[node]
                            Rx = (bx[btri[i,k]] + bx[btri[i,l]])/2 - px
                            Ry = (by[btri[i,k]] + by[btri[i,l]])/2 - py
                            Ex = Ex + mesh.nodeSource[node]*pc*Rx/(np.sqrt(Rx**2 + Ry**2))**2
                            Ey = Ey + mesh.nodeSource[node]*pc*Ry/(np.sqrt(Rx**2 + Ry**2))**2

                    else:
                        nodes = np.concatenate((domain.boundaryList, domain.freeList)).astype(int)
                        for node in nodes:
                            px = x[node]
                            py = y[node]
                            Rx = (bx[btri[i,k]] + bx[btri[i,l]])/2 - px
                            Ry = (by[btri[i,k]] + by[btri[i,l]])/2 - py
                            Ex = Ex + mesh.nodeSource[node]*pc*Rx/(np.sqrt(Rx**2 + Ry**2))**2
                            Ey = Ey + mesh.nodeSource[node]*pc*Ry/(np.sqrt(Rx**2 + Ry**2))**2

            rx = bx[btri[i,l]] - bx[btri[i,k]]
            ry = by[btri[i,l]] - by[btri[i,k]]
            Er = np.dot([Ex[0], Ey[0]], [rx, ry]/np.sqrt(rx**2+ry**2))
            A[btri[i,k], btri[i,k]] = A[btri[i,k], btri[i,k]] +1
            A[btri[i,k], btri[i,l]] = -1
            b[btri[i,k]]           = b[btri[i,k]] + Er*np.sqrt(rx**2+ry**2)

            # 2-0 edge
            k = 2
            l = 0
            Ex = 0
            Ey = 0

            for domain in initialDomains:
                if domain.source != 0:
                    if domain.shape == "point" or domain.shape == "line":
                        nodes = domain.boundaryList
                        for node in nodes:
                            px = x[node]
                            py = y[node]
                            Rx = (bx[btri[i,k]] + bx[btri[i,l]])/2 - px
                            Ry = (by[btri[i,k]] + by[btri[i,l]])/2 - py
                            Ex = Ex + mesh.nodeSource[node]*pc*Rx/(np.sqrt(Rx**2 + Ry**2))**2
                            Ey = Ey + mesh.nodeSource[node]*pc*Ry/(np.sqrt(Rx**2 + Ry**2))**2

                    else:
                        nodes = np.concatenate((domain.boundaryList, domain.freeList)).astype(int)
                        for node in nodes:
                            px = x[node]
                            py = y[node]
                            Rx = (bx[btri[i,k]] + bx[btri[i,l]])/2 - px
                            Ry = (by[btri[i,k]] + by[btri[i,l]])/2 - py
                            Ex = Ex + mesh.nodeSource[node]*pc*Rx/(np.sqrt(Rx**2 + Ry**2))**2
                            Ey = Ey + mesh.nodeSource[node]*pc*Ry/(np.sqrt(Rx**2 + Ry**2))**2

            rx = bx[btri[i,l]] - bx[btri[i,k]]
            ry = by[btri[i,l]] - by[btri[i,k]]
            Er = np.dot([Ex[0], Ey[0]], [rx, ry]/np.sqrt(rx**2+ry**2))
            A[btri[i,k], btri[i,k]] = A[btri[i,k], btri[i,k]] +1
            A[btri[i,k], btri[i,l]] = -1
            b[btri[i,k]]           = b[btri[i,k]] + Er*np.sqrt(rx**2+ry**2)

        print()
        self.Aab = A
        self.bab = b
        mesh.bx  = bx
        mesh.by  = by
        mesh.btri= btri

    def getIterBoundaryMatrices(self, mesh, initialDomains, bt):
        """
        gets approximate matrices on boundary for iterative update

        Parameters
        ----------
        mesh : object
            the mesh object
        initialDomains : list
            list of geometries
        bt : list

        Returns
        -------
        None
        """

        x = mesh.rx
        y = mesh.ry
        tri = mesh.tri
        u0 = self.physicalConstant
        pc = self.proportionalityConstant

        # Boundary nodes
        gc = initialDomains[0].childDomains
        lbn = initialDomains[0].boundaryList
        lbn = np.append([lbn], [initialDomains[0].freeList])
        lbn = np.append([lbn], [initialDomains[gc[0]].boundaryList])
        lbn = lbn.astype(int)
        bx = x[lbn].copy()
        by = y[lbn].copy()

        # Boundary elements
        belements = initialDomains[0].elementList
        btri = tri[belements].copy()
        bne = len(btri[:,0])
        bnn = len(bx)

        # Mapping btri and tri

        for j in range(bne):
            for k in range(3):
                gn = btri[j,k]

                for l in range(bnn):
                    if gn == lbn[l]:
                        btri[j,k] = l
                        break
        
        # Initializing matrices
        A = np.zeros([bnn,bnn])
        b = np.zeros([bnn,1])

        # For every boundary element
        for i in range(bne):

            print("Percentage of boundary elements covered", i/bne*100, end='\r')

            # 0-1 edge
            k = 0
            l = 1
            Ex = 0
            Ey = 0

            for domain in initialDomains:
                if domain.source != 0 or domain.material != 1:
                    if domain.shape == "point" or domain.shape == "line":
                        nodes = domain.boundaryList
                        for node in nodes:
                            px = x[node]
                            py = y[node]
                            Rx = (bx[btri[i,k]] + bx[btri[i,l]])/2 - px
                            Ry = (by[btri[i,k]] + by[btri[i,l]])/2 - py
                            Ex = Ex + bt[node]/u0*pc*Rx/(np.sqrt(Rx**2 + Ry**2))**2
                            Ey = Ey + bt[node]/u0*pc*Ry/(np.sqrt(Rx**2 + Ry**2))**2

                    else:
                        nodes = np.concatenate((domain.boundaryList, domain.freeList)).astype(int)
                        for node in nodes:
                            px = x[node]
                            py = y[node]
                            Rx = (bx[btri[i,k]] + bx[btri[i,l]])/2 - px
                            Ry = (by[btri[i,k]] + by[btri[i,l]])/2 - py
                            Ex = Ex + bt[node]/u0*pc*Rx/(np.sqrt(Rx**2 + Ry**2))**2
                            Ey = Ey + bt[node]/u0*pc*Ry/(np.sqrt(Rx**2 + Ry**2))**2

                
            rx = bx[btri[i,l]] - bx[btri[i,k]]
            ry = by[btri[i,l]] - by[btri[i,k]]
            Er = np.dot([Ex[0], Ey[0]], [rx, ry]/np.sqrt(rx**2+ry**2))
            A[btri[i,k], btri[i,k]] = A[btri[i,k], btri[i,k]] +1
            A[btri[i,k], btri[i,l]] = -1
            b[btri[i,k]]           = b[btri[i,k]] + Er*np.sqrt(rx**2+ry**2)

            # 1-2 edge
            k = 1
            l = 2
            Ex = 0
            Ey = 0

            for domain in initialDomains:
                if domain.source != 0 or domain.material != 1:
                    if domain.shape == "point" or domain.shape == "line":
                        nodes = domain.boundaryList
                        for node in nodes:
                            px = x[node]
                            py = y[node]
                            Rx = (bx[btri[i,k]] + bx[btri[i,l]])/2 - px
                            Ry = (by[btri[i,k]] + by[btri[i,l]])/2 - py
                            Ex = Ex + bt[node]/u0*pc*Rx/(np.sqrt(Rx**2 + Ry**2))**2
                            Ey = Ey + bt[node]/u0*pc*Ry/(np.sqrt(Rx**2 + Ry**2))**2

                    else:
                        nodes = np.concatenate((domain.boundaryList, domain.freeList)).astype(int)
                        for node in nodes:
                            px = x[node]
                            py = y[node]
                            Rx = (bx[btri[i,k]] + bx[btri[i,l]])/2 - px
                            Ry = (by[btri[i,k]] + by[btri[i,l]])/2 - py
                            Ex = Ex + bt[node]/u0*pc*Rx/(np.sqrt(Rx**2 + Ry**2))**2
                            Ey = Ey + bt[node]/u0*pc*Ry/(np.sqrt(Rx**2 + Ry**2))**2

            rx = bx[btri[i,l]] - bx[btri[i,k]]
            ry = by[btri[i,l]] - by[btri[i,k]]
            Er = np.dot([Ex[0], Ey[0]], [rx, ry]/np.sqrt(rx**2+ry**2))
            A[btri[i,k], btri[i,k]] = A[btri[i,k], btri[i,k]] +1
            A[btri[i,k], btri[i,l]] = -1
            b[btri[i,k]]           = b[btri[i,k]] + Er*np.sqrt(rx**2+ry**2)

            # 2-0 edge
            k = 2
            l = 0
            Ex = 0
            Ey = 0

            for domain in initialDomains:
                if domain.source != 0 or domain.material != 1:
                    if domain.shape == "point" or domain.shape == "line":
                        nodes = domain.boundaryList
                        for node in nodes:
                            px = x[node]
                            py = y[node]
                            Rx = (bx[btri[i,k]] + bx[btri[i,l]])/2 - px
                            Ry = (by[btri[i,k]] + by[btri[i,l]])/2 - py
                            Ex = Ex + bt[node]/u0*pc*Rx/(np.sqrt(Rx**2 + Ry**2))**2
                            Ey = Ey + bt[node]/u0*pc*Ry/(np.sqrt(Rx**2 + Ry**2))**2

                    else:
                        nodes = np.concatenate((domain.boundaryList, domain.freeList)).astype(int)
                        for node in nodes:
                            px = x[node]
                            py = y[node]
                            Rx = (bx[btri[i,k]] + bx[btri[i,l]])/2 - px
                            Ry = (by[btri[i,k]] + by[btri[i,l]])/2 - py
                            Ex = Ex + bt[node]/u0*pc*Rx/(np.sqrt(Rx**2 + Ry**2))**2
                            Ey = Ey + bt[node]/u0*pc*Ry/(np.sqrt(Rx**2 + Ry**2))**2

            rx = bx[btri[i,l]] - bx[btri[i,k]]
            ry = by[btri[i,l]] - by[btri[i,k]]
            Er = np.dot([Ex[0], Ey[0]], [rx, ry]/np.sqrt(rx**2+ry**2))
            A[btri[i,k], btri[i,k]] = A[btri[i,k], btri[i,k]] +1
            A[btri[i,k], btri[i,l]] = -1
            b[btri[i,k]]           = b[btri[i,k]] + Er*np.sqrt(rx**2+ry**2)

        print()
        self.Aib = A
        self.bib = b
        mesh.bx  = bx
        mesh.by  = by
        mesh.btri= btri

    def getMiniLinNumericalMatrix(self, minix, miniy, miniTri):
        """
        gets linear numerical matrices

        Parameters
        ----------
        minix : list
            x positions of minimesh
        miniy : list
            y positions of minimesh
        miniTri : list
            connectivity matrix of minimesh
        Returns
        -------
        None
        """

        x = minix
        y = miniy
        tri = miniTri
        nn = len(x)
        A = np.zeros((nn,nn))

        count = 0
        ne = len(tri[:,0])
        for i in range(ne):
            print("percentage of linear elements covered", count/ne*100, end='\r')
            count = count +1
            ex = np.array([x[tri[i,0]], x[tri[i,1]], x[tri[i,2]]])
            ey = np.array([y[tri[i,0]], y[tri[i,1]], y[tri[i,2]]])
            p = np.array([ey[1]-ey[2], ey[2]-ey[0], ey[0]-ey[1]])
            q = np.array([ex[2]-ex[1], ex[0]-ex[2], ex[1]-ex[0]])
            ea = 0.5*abs(p[1]*q[2]-q[1]*p[2])
            P = np.zeros([len(p),len(p)])
            Q = P.copy()
            for j in range(len(p)):
                P[j] = p*p[j]
                Q[j] = q*q[j]

            a = (P+Q)/(4*ea)
            for k in range(3):
                for j in range(3):
                    A[tri[i,k],tri[i,j]] = A[tri[i,k],tri[i,j]].copy() + a[k,j]

        # print()
        self.miniAn = A.copy()

    def findVectorFields(self, mesh, solver):
        """
        finds gradient vector fields

        Parameters
        ----------
        mesh : object
            the mesh object
        solver : object
            the solver object

        Returns
        -------
        None
        """
        
        x = mesh.rx
        y = mesh.ry
        u = solver.u.copy()

        Ex = np.zeros([len(x),1])
        Ey = np.zeros([len(x),1])
        farnodeAttNames = mesh.farnodeAttNames.copy()
        neiNodesTn = getattr(mesh, farnodeAttNames[0])

        for i in range(len(x)):
            tNodes = neiNodesTn[i]
            for node in tNodes:
                rx = x[node]-x[i]
                ry = y[node]-y[i]
                # print(rx, ry, (np.sqrt(rx**2 + ry**2))**2)
                Ex[i] = Ex[i] + (u[i] - u[node]) * rx / (np.sqrt(rx**2 + ry**2))**2
                Ey[i] = Ey[i] + (u[i] - u[node]) * ry / (np.sqrt(rx**2 + ry**2))**2
        
        thr = 90
        R = np.array([[np.cos(np.radians(thr)), -np.sin(np.radians(thr))],
        [np.sin(np.radians(thr)), np.cos(np.radians(thr))]])
        
        for i in range(len(Ex)):
            Ex[i], Ey[i] = np.dot(R,np.array([Ex[i], Ey[i]]))[0], np.dot(R,np.array([Ex[i], Ey[i]]))[1]
        
        solver.Ex = Ex
        solver.Ey = Ey



class Domain:
    """
    A class to represent a geometry.
    
    ...

    Attributes
    ----------
    width : float
        width of the geometry
    height : float
        height of the geometry
    position : list
        position of the most significant point of the geometry
    relativeMeshDensity : float
        relative mesh density with respect to the global geometry
    angle : float
        angle of rotation of the geometry around the most significant point of the geometry
    isChild : boolean
        states if a geometry is a child of another geometry
    isParent : boolean
        states if a geometry is a parent of another geometry
    childDomains : list
        contains list of child domains in the geometry
    parentDomains : int
        contains the number of parent domain of current geometry
    non : int
        number of nodes to be forced manually along the width of the most significant perimeter

    Methods
    -------
    calculateRelativeWidth(gWidth):
        calculates relative width of a geometry with respects to the global geometry width.
    calculateWidthNodes(gNon):
        assigns number of nodes along the width of a geometry according to non or relative mesh density.
    """
    
    def __init__(self, width, height, position, relativeMeshDensity, angle, isChild, isParent, childDomains, parentDomains, non):
        """
        Constructs all the necessary attributes for the geometry object.

        Parameters
        ----------
        width : float
            width of the geometry
        height : float
            height of the geometry
        position : list
            position of the most significant point of the geometry
        relativeMeshDensity : float
            relative mesh density with respect to the global geometry
        angle : float
            angle of rotation of the geometry around the most significant point of the geometry
        isChild : boolean
            states if a geometry is a child of another geometry
        isParent : boolean
            states if a geometry is a parent of another geometry
        childDomains : list
            contains list of child domains in the geometry
        parentDomains : int
            contains the number of parent domain of current geometry
        non : int
            number of nodes to be forced manually along the width of the most significant perimeter
        """

        self.width = width
        self.height = height
        self.position = position
        self.relativeMeshDensity = relativeMeshDensity
        self.angle = angle
        self.isChild = isChild
        self.isParent = isParent
        self.childDomains = childDomains
        self.parentDomains = parentDomains
        self.non = non

    def calculateRelativeWidth(self, gWidth):
        """
        calculates relative width of a geometry with respects to the global geometry width.

        Parameters
        ----------
        gwidth : float
            width og gloabl geometry

        Returns
        -------
        None
        """

        self.relativeWidth = self.width/gWidth

    def calculateWidthNodes(self, gNon):
        """
        assigns number of nodes along the width of a geometry according to non or relative mesh density.

        Parameters
        ----------
        gNon : int
            number of nodes on the width of global boundary

        Returns
        -------
        None
        """
        
        if self.non == 0:
            self.widthNodes = round(self.relativeWidth*self.relativeMeshDensity[0]*gNon+1)
        else:
            self.widthNodes = self.non

    



import numpy as np
import random

from . import Domain

class Ellipse(Domain.Domain):
    def __init__(self, width, height, position, relativeMeshDensity, angle, isChild, isParent, childDomains, parentDomains, non):
        """
        Constructs all the necessary attributes for the geometry object.

        Parameters
        ----------
        shape: str
            the name of the shape of geometry
        """
        
        Domain.Domain.__init__(self, width, height, position, relativeMeshDensity, angle, isChild, isParent, childDomains, parentDomains, non)
        self.shape = "ellipse"

    def show(self, plt, lW):
        """
        adds the current geometry to plot

        Parameters
        ----------
        plt : object
            matplotlib object
        lw : float
            width of line stroke

        Returns
        -------
        None
        """

        # Generating boundary nodes for plotting
        a = self.width/2
        b = self.height/2
        th = np.linspace(0,2*np.pi,100)
        x = a*np.cos(th)
        y = b*np.sin(th)
        v0 = np.array([x,y])
        # Rotate the boundary points
        thr = self.angle
        R = np.array([[np.cos(np.radians(thr)), -np.sin(np.radians(thr))],
        [np.sin(np.radians(thr)), np.cos(np.radians(thr))]])
        vr = np.dot(R,v0)
        # Reposition the boundary points
        position = np.array(self.position)
        v = position +  np.transpose(vr)
        # Plot the boundary points
        plt.plot(v[:,0], v[:,1], 'k', lw = lW)

    def calculateArea(self):
        """
        calculates area.

        Parameters
        ----------
        None

        Returns
        -------
        None
        """

        self.area = np.pi/4*self.width*self.height

    def calculateMaxFreeNodes(self, domains):
        """
        calculates maximum number of free nodes allowed

        Parameters
        ----------
        domains : list
            list of all geometry objects

        Returns
        -------
        None
        """

        ab = domains[0].non
        a = domains[0].width
        b = domains[0].height
        mtn = round(1.2*(ab-2)**2*np.pi/4*b/a)

        area = self.area
        if self.isParent:
            for childDomain in self.childDomains:
                area = area - domains[childDomain].area

        ga = domains[0].area
        rd = self.relativeMeshDensity[1]
        self.maxFreeNodes = round(mtn*area/ga*rd**2)

    def genVertices(self, plt, lW, markerShape):
        """
        gets vertices of the geometry.

        Parameters
        ----------
        plt : object
            matplotlib object
        lw : float
            width of line stroke
        markerShape : char
            shape of marker

        Returns
        -------
        None
        """

        # Generating vertices
        a = self.width/2
        v0 = np.array([a, 0])
        v0 = np.transpose(v0)
        # Rotate the boundary points
        thr = self.angle
        R = np.array([[np.cos(np.radians(thr)), -np.sin(np.radians(thr))],
        [np.sin(np.radians(thr)), np.cos(np.radians(thr))]])
        vr = np.dot(R,v0)
        # Reposition the boundary points
        position = np.array(self.position)
        v = position +  np.transpose(vr)
        self.verticesX = v[0]
        self.verticesY = v[1]
        # Plot the vertices
        plt.scatter(self.verticesX, self.verticesY, c=None, marker=markerShape, s = lW*1000)

    def genBoundaryNodes(self, plt, lW, markerShape):
        """
        gets boundary nodes.

        Parameters
        ----------
        plt : object
            matplotlib object
        lw : float
            width of line stroke
        markerShape : char
            shape of marker

        Returns
        -------
        None
        """

        # Generate boundary points
        a = self.width/2
        b = self.height/2
        per = np.pi*((3*(a+b))-np.sqrt((3*a+b)*(a+3*b)))
        nbn = round(per/(2*a)*self.widthNodes)
        th  = np.linspace(0,2*np.pi,nbn)
        x = a*np.cos(th)[0:-1]
        y = b*np.sin(th)[0:-1]
        v0 = np.array([x,y])
        # Rotate the boundary points
        thr = self.angle
        R = np.array([[np.cos(np.radians(thr)), -np.sin(np.radians(thr))],
        [np.sin(np.radians(thr)), np.cos(np.radians(thr))]])
        vr = np.dot(R,v0)
        # Reposition the boundary points
        position = np.array(self.position)
        v = position +  np.transpose(vr)
        b = np.transpose(v)
        self.boundaryNodesX = b[0]
        self.boundaryNodesY = b[1]
        plt.scatter(self.boundaryNodesX, self.boundaryNodesY, c=None, marker=markerShape, s = lW*1000)

    def genRandFreeNodes(self, initialDomains, plt, lW, markerShape):
        """
        gets random free nodes inside a domain.

        Parameters
        ----------
        initialDomains : list
            list of all geometry objects
        plt : object
            matplotlib object
        lw : float
            width of line stroke
        markerShape : char
            shape of marker

        Returns
        -------
        None
        """

        hbyw = self.height/self.width
        mnf = self.maxFreeNodes
        thr = self.angle
        R = np.array([[np.cos(np.radians(thr)), -np.sin(np.radians(thr))],
        [np.sin(np.radians(thr)), np.cos(np.radians(thr))]])

        fix = self.boundaryNodesX
        fiy = self.boundaryNodesY
        h = self.width/self.widthNodes

        if self.isParent: # Add child nodes as fixed nodes
            for cd in self.childDomains:
                fix = np.append(fix, initialDomains[cd].allNodesX)
                fiy = np.append(fiy, initialDomains[cd].allNodesY)

        nop = 0 # Initializing no. of points inserted
        count = 0 # Count for while loop
        fx = np.array([0]) # Initializing free node positions
        fy = np.array([0])

        if hbyw >= 1:
            maxit = 1000*mnf*hbyw*4/np.pi

            while nop < mnf and count < maxit:
                # Get random points
                xr = random.uniform(0,1) - 0.5
                yr = random.uniform(0,1) - 0.5

                if (((xr**2)/((0.5/hbyw)**2))+((yr**2)/((0.5)**2))) <= 1:
                    # Scale the random point
                    xr = xr*self.height
                    yr = yr*self.height
                    # Rotate the random point
                    p = np.array([xr,yr])
                    p = np.dot(R,p)
                    # Reposition the point
                    p = self.position + p
                    minDis = min(np.sqrt((np.append(fix, fx)-p[0])**2+(np.append(fiy, fy)-p[1])**2))

                    if minDis > h:
                        fx = np.append(fx, p[0])
                        fy = np.append(fy, p[1])

                    count = count +1

        else:
            maxit = 1000*mnf/hbyw*4/np.pi

            while nop < mnf and count < maxit:
                # Get random points
                xr = random.uniform(0,1) - 0.5
                yr = random.uniform(0,1) - 0.5

                if (((xr**2)/((0.5)**2))+((yr**2)/((0.5*hbyw)**2))) <= 1:
                    # Scale the random point
                    xr = xr*self.width
                    yr = yr*self.width
                    # Rotate the random point
                    p = np.array([xr,yr])
                    p = np.dot(R,p)
                    # Reposition the point
                    p = self.position + p
                    minDis = min(np.sqrt((np.append(fix, fx)-p[0])**2+(np.append(fiy, fy)-p[1])**2))

                    if minDis > h:
                        fx = np.append(fx, p[0])
                        fy = np.append(fy, p[1])

                    count = count +1

        self.freeNodesX = fx[1:None]
        self.freeNodesY = fy[1:None]
        self.allNodesX = np.concatenate((self.boundaryNodesX, self.freeNodesX))
        self.allNodesY = np.concatenate((self.boundaryNodesY, self.freeNodesY))

        for k in self.childDomains:
            if initialDomains[k].shape == "point":
                self.allNodesX = np.concatenate((self.allNodesX, [initialDomains[k].allNodesX]))
                self.allNodesY = np.concatenate((self.allNodesY, [initialDomains[k].allNodesY]))
            else:
                self.allNodesX = np.concatenate((self.allNodesX, initialDomains[k].allNodesX))
                self.allNodesY = np.concatenate((self.allNodesY, initialDomains[k].allNodesY))

        plt.scatter(self.freeNodesX, self.freeNodesY, c=None, marker=markerShape, s = lW*1000)

    def genTriFreeNodes(self, initialDomains, plt, lW, markerShape):
        """
        gets uniformly distributed triangular free nodes inside a domain.

        Parameters
        ----------
        initialDomains : list
            list of all geometry objects
        plt : object
            matplotlib object
        lw : float
            width of line stroke
        markerShape : char
            shape of marker

        Returns
        -------
        None
        """

        hbyw = self.height/self.width
        thr = self.angle
        R = np.array([[np.cos(np.radians(thr)), -np.sin(np.radians(thr))],
        [np.sin(np.radians(thr)), np.cos(np.radians(thr))]])

        fix = self.boundaryNodesX
        fiy = self.boundaryNodesY
        h = self.width/self.widthNodes/1.25*self.relativeMeshDensity[0]/self.relativeMeshDensity[1]

        if self.isParent: # Add child nodes as fixed nodes
            for cd in self.childDomains:
                fix = np.append(fix, initialDomains[cd].allNodesX)
                fiy = np.append(fiy, initialDomains[cd].allNodesY)

        fx = np.array([]) # Initializing free node positions
        fy = np.array([])

        if hbyw >= 1:

            wn = round(self.relativeWidth*self.relativeMeshDensity[1]*initialDomains[0].non)
            wn = round(hbyw*wn*0.8)
            xh = np.linspace(-0.788, 0.788, 2*wn-1)
            yh = 0*xh

            for i in range(wn-1):
                xd = np.linspace(-0.788/(2*wn-2)*(2*wn-i-3), 0.788/(2*wn-2)*(2*wn-i-3), 2*wn-i-2)
                xh = np.concatenate([xh, xd, xd])
                yp = 0*xd + 1.576/(2*wn-2)*np.sqrt(3)/2*(i+1)
                yn = -yp
                yh = np.concatenate([yh, yp, yn])

            for i in range(len(xh)):

                if (((xh[i]**2)/((0.5/hbyw)**2))+((yh[i]**2)/((0.5)**2))) < 1:
                    
                    xr = xh[i]
                    yr = yh[i]
                    # Scale the random point
                    xr = xr*self.height
                    yr = yr*self.height
                    # Rotate the random point
                    p = np.array([xr,yr])
                    p = np.dot(R,p)
                    # Reposition the point
                    p = self.position + p
                    minDis = min(np.sqrt((np.append(fix, fx)-p[0])**2+(np.append(fiy, fy)-p[1])**2))

                    if minDis > h:
                        fx = np.append(fx, p[0])
                        fy = np.append(fy, p[1])

        else:
            wn = round(self.relativeWidth*self.relativeMeshDensity[1]*initialDomains[0].non*0.8)
            xh = np.linspace(-0.788, 0.788, 2*wn-1)
            yh = 0*xh

            for i in range(wn-1):
                xd = np.linspace(-0.788/(2*wn-2)*(2*wn-i-3), 0.788/(2*wn-2)*(2*wn-i-3), 2*wn-i-2)
                xh = np.concatenate([xh, xd, xd])
                yp = 0*xd + 1.576/(2*wn-2)*np.sqrt(3)/2*(i+1)
                yn = -yp
                yh = np.concatenate([yh, yp, yn])

            for i in range(len(xh)):

                if (((xh[i]**2)/((0.5)**2))+((yh[i]**2)/((0.5*hbyw)**2))) < 1:
                    
                    xr = xh[i]
                    yr = yh[i]
                    # Scale the random point
                    xr = xr*self.width
                    yr = yr*self.width
                    # Rotate the random point
                    p = np.array([xr,yr])
                    p = np.dot(R,p)
                    # Reposition the point
                    p = self.position + p
                    minDis = min(np.sqrt((np.append(fix, fx)-p[0])**2+(np.append(fiy, fy)-p[1])**2))

                    if minDis > h:
                        fx = np.append(fx, p[0])
                        fy = np.append(fy, p[1])

        self.freeNodesX = fx[1:None]
        self.freeNodesY = fy[1:None]
        self.allNodesX = np.concatenate((self.boundaryNodesX, self.freeNodesX))
        self.allNodesY = np.concatenate((self.boundaryNodesY, self.freeNodesY))

        for k in self.childDomains:
            if initialDomains[k].shape == "point":
                self.allNodesX = np.concatenate((self.allNodesX, [initialDomains[k].allNodesX]))
                self.allNodesY = np.concatenate((self.allNodesY, [initialDomains[k].allNodesY]))
            else:
                self.allNodesX = np.concatenate((self.allNodesX, initialDomains[k].allNodesX))
                self.allNodesY = np.concatenate((self.allNodesY, initialDomains[k].allNodesY))

        plt.scatter(self.freeNodesX, self.freeNodesY, c=None, marker=markerShape, s = lW*1000)





    


# Import packages
import joblib

import DssPyLib.Mesh as Mesh
import DssPyLib.PlotUtils as PlotUtils
import DssPyLib.Solvers.Solve as Solve


def nodes(meshingMethod = "tri", figSize = "Default", showPlot = False):
    """
    generates nodes.

    Parameters
    ----------
    meshingMethod : str
        method of meshing
    figSize : list
        size of figure
    showPlot : boolean
        show plot if true

    Returns
    -------
    None
    """

    initialDomains = joblib.load("Data/initialDomains.sav")
    Mesh.generateNodes(initialDomains, meshingMethod, figSize, showPlot)
    joblib.dump(initialDomains, "Data/initialDomains.sav")
    
def mesh(meshNumbering = "unNumbered", figSize = "Default", showPlot = False):
    """
    generates mesh.

    Parameters
    ----------
    meshNumbering : str
        string mentioning numbered or unNumbered mesh
    figSize : list
        size of figure
    showPlot : boolean
        show plot if true

    Returns
    -------
    None
    """

    initialDomains = joblib.load("Data/initialDomains.sav")
    mesh = Mesh.createMesh(initialDomains)
    joblib.dump(initialDomains, "Data/initialDomains.sav")
    joblib.dump(mesh, "Data/mesh.sav")
    PlotUtils.plotMesh(initialDomains, mesh, "Figures/2_Initial_mesh.png", figSize, showPlot, meshNumbering)

def relaxedMesh(meshNumbering = "unNumbered", figSize = "Default", showPlot = False):
    """
    relaxes mesh.

    Parameters
    ----------
    meshNumbering : str
        string mentioning numbered or unNumbered mesh
    figSize : list
        size of figure
    showPlot : boolean
        show plot if true

    Returns
    -------
    None
    """

    initialDomains = joblib.load("Data/initialDomains.sav")
    mesh = joblib.load("Data/mesh.sav")
    Mesh.relaxMesh(initialDomains, mesh)
    joblib.dump(initialDomains, "Data/initialDomains.sav")
    joblib.dump(mesh, "Data/mesh.sav")
    PlotUtils.plotMesh(initialDomains, mesh, "Figures/2_Relaxed_mesh.png", figSize, showPlot, meshNumbering)

def assigedSources(colorbarPosition, figSize, showPlot):
    """
    assigns sources.

    Parameters
    ----------
    colorbarPosition : str
        position of color bar
    figSize : list
        size of figure
    showPlot : boolean
        show plot if true

    Returns
    -------
    None
    """

    initialDomains = joblib.load("Data/initialDomains.sav")
    mesh = joblib.load("Data/mesh.sav")
    solver = joblib.load("Data/solver.sav")
    Mesh.assignSources(initialDomains, mesh, solver)
    joblib.dump(mesh, "Data/mesh.sav")
    PlotUtils.plotContour(initialDomains, mesh, solver, "Source", "Figures/3_Sources.png", colorbarPosition, figSize, showPlot)

def assigedMaterials(colorbarPosition, figSize, showPlot):
    """
    assigns materials.

    Parameters
    ----------
    colorbarPosition : str
        position of color bar
    figSize : list
        size of figure
    showPlot : boolean
        show plot if true

    Returns
    -------
    None
    """

    initialDomains = joblib.load("Data/initialDomains.sav")
    mesh = joblib.load("Data/mesh.sav")
    solver = joblib.load("Data/solver.sav")
    Mesh.assignMaterials(initialDomains, mesh)
    joblib.dump(mesh, "Data/mesh.sav")
    PlotUtils.plotContour(initialDomains, mesh, solver, "Material", "Figures/3_Materials.png", colorbarPosition, figSize, showPlot)

def neighbourInformation(truncationNumber = 5):
    """
    generates neighbour information.

    Parameters
    ----------
    truncationNumber : int
        truncation number    

    Returns
    -------
    None
    """

    mesh = joblib.load("Data/mesh.sav")
    Mesh.getNeighbourInformation(mesh, truncationNumber)
    joblib.dump(mesh, "Data/mesh.sav")

def solution(solutionName = "numerical", solverMethod = "Scipy", figSize = "Default", colorbarPosition = "right", showPlot = False, withField = False, withDifference = False):
    """
    generates solution.

    Parameters
    ----------
    solutionName : str
        name of solution
    solverMethod : str
        name of solver
    figSize : list
        size of figure
    colorbarPosition : str
        position of colorbar
    showPlot : boolean
        show plot if true
    withField : boolean
        show vector field if true
    withDifference : boolean
        show difference in solution if true

    Returns
    -------
    None
    """

    initialDomains = joblib.load("Data/initialDomains.sav")
    mesh = joblib.load("Data/mesh.sav")
    solver = joblib.load("Data/solver.sav")
    Solve.getMatrices(initialDomains, mesh, solver, solutionName)
    Solve.solve(initialDomains, mesh, solver, solutionName, solverMethod)
    showSolution(mesh, solver, solutionName, figSize, colorbarPosition, showPlot, withField)
    if withDifference:
        showDifference(mesh, solver, solutionName, figSize, showPlot)
    joblib.dump(mesh, "Data/mesh.sav")
    joblib.dump(solver, "Data/solver.sav")

def dssSolution(solutionName = "numerical", truncationNumber = 5, solverMethod = "Scipy", figSize = "Default", colorbarPosition = "right", showPlot = False, withField = False, withDifference = False):
    """
    generates dss solution.

    Parameters
    ----------
    solutionName : str
        name of solution
    truncationNumber : int
        truncation number 
    solverMethod : str
        name of solver
    figSize : list
        size of figure
    colorbarPosition : str
        position of colorbar
    showPlot : boolean
        show plot if true
    withField : boolean
        show vector field if true
    withDifference : boolean
        show difference in solution if true

    Returns
    -------
    None
    """

    initialDomains = joblib.load("Data/initialDomains.sav")
    mesh = joblib.load("Data/mesh.sav")
    solver = joblib.load("Data/solver.sav")
    Solve.getDistributedSources(initialDomains, mesh, solver, truncationNumber)
    Solve.getMatrices(initialDomains, mesh, solver, solutionName)
    Solve.solveDSS(initialDomains, mesh, solver, solutionName, solverMethod)
    showDSSSolution(mesh, solver, solutionName, figSize, colorbarPosition, showPlot, withField)
    showDistributedSources(initialDomains, mesh, solver, solutionName, colorbarPosition, figSize, showPlot)
    if withDifference:
        showDSSDifference(mesh, solver, solutionName, figSize, showPlot)
    joblib.dump(mesh, "Data/mesh.sav")
    joblib.dump(solver, "Data/solver.sav")

def showSolution(mesh, solver, solutionName, figSize, colorbarPosition, showPlot, withField):
    """
    shows solution.

    Parameters
    ----------
    mesh : object
        mesh object
    solver : object
        solver object
    solutionName : str
        name of solution
    figSize : list
        size of figure
    colorbarPosition : str
        position of colorbar
    showPlot : boolean
        show plot if true
    withField : boolean
        show vector field if true

    Returns
    -------
    None
    """

    domains = joblib.load("Data/initialDomains.sav")
    if (solutionName == "approximate"):
        if (withField):
            PlotUtils.plotContourField(domains, "Figures/4_Approximate_solution", mesh, solver, solutionName, figSize, colorbarPosition, showPlot)
        else:
            solution = solver.ua
            PlotUtils.plotSurf(domains, "Figures/4_Approximate_solution", mesh, solver, solution, solutionName, figSize, showPlot)
    elif (solutionName == "numerical"):
        if (withField):
            PlotUtils.plotContourField(domains, "Figures/4_Numerical_solution", mesh, solver, solutionName, figSize, colorbarPosition, showPlot)
        else:
            solution = solver.un
            PlotUtils.plotSurf(domains, "Figures/4_Numerical_solution", mesh, solver, solution, solutionName, figSize, showPlot)
    elif (solutionName == "numericalApproxBound"):
        if (withField):
            PlotUtils.plotContourField(domains, "Figures/4_Numerical_solution_approx_boundary", mesh, solver, solutionName, figSize, colorbarPosition, showPlot)
        else:
            solution = solver.unab
            PlotUtils.plotSurf(domains, "Figures/4_Numerical_solution_approx_boundary", mesh, solver, solution, solutionName, figSize, showPlot)
    elif (solutionName == "numericalIterBound"):
        if (withField):
            PlotUtils.plotContourField(domains, "Figures/4_Numerical_solution_iter_boundary", mesh, solver, solutionName, figSize, colorbarPosition, showPlot)
        else:
            solution = solver.unib
            PlotUtils.plotSurf(domains, "Figures/4_Numerical_solution_iter_boundary", mesh, solver, solution, solutionName, figSize, showPlot)

def showDSSSolution(mesh, solver, solutionName, figSize, colorbarPosition, showPlot, withField):
    """
    shows dss solution.

    Parameters
    ----------
    mesh : object
        mesh object
    solver : object
        solver object
    solutionName : str
        name of solution
    figSize : list
        size of figure
    colorbarPosition : str
        position of colorbar
    showPlot : boolean
        show plot if true
    withField : boolean
        show vector field if true

    Returns
    -------
    None
    """

    domains = joblib.load("Data/initialDomains.sav")
    if (solutionName == "approximate"):
        print("DSS does not support approximate solution")
    elif (solutionName == "numerical"):
        if (withField):
            PlotUtils.plotContourField(domains, "Figures/4_Numerical_DSS_solution", mesh, solver, solutionName, figSize, colorbarPosition, showPlot)
        else:
            solution = solver.un
            PlotUtils.plotSurf(domains, "Figures/4_Numerical_DSS_solution", mesh, solver, solution, solutionName, figSize, showPlot)
    elif (solutionName == "numericalApproxBound"):
        if (withField):
            PlotUtils.plotContourField(domains, "Figures/4_Numerical_DSS_solution_approx_boundary", mesh, solver, solutionName, figSize, colorbarPosition, showPlot)
        else:
            solution = solver.unab
            PlotUtils.plotSurf(domains, "Figures/4_Numerical_DSS_solution_approx_boundary", mesh, solver, solution, solutionName, figSize, showPlot)
    elif (solutionName == "numericalIterBound"):
        if (withField):
            PlotUtils.plotContourField(domains, "Figures/4_Numerical_DSS_solution_iter_boundary", mesh, solver, solutionName, figSize, colorbarPosition, showPlot)
        else:
            solution = solver.unib
            PlotUtils.plotSurf(domains, "Figures/4_Numerical_DSS_solution_iter_boundary", mesh, solver, solution, solutionName, figSize, showPlot)

def showDifference(mesh, solver, solutionName, figSize, showPlot):
    """
    shows difference in solution.

    Parameters
    ----------
    mesh : object
        mesh object
    solver : object
        solver object
    solutionName : str
        name of solution
    figSize : list
        size of figure
    showPlot : boolean
        show plot if true

    Returns
    -------
    None
    """

    domains = joblib.load("Data/initialDomains.sav")
    if (solutionName == "approximate"):
        print("Difference is not supported for approximate solution")
    elif (solutionName == "numerical"):
        appxsolution = solver.ua
        solution = solver.un
        PlotUtils.plotDifference(domains, "Figures/5_Difference_in_Numerical_solution", mesh, solver, appxsolution, solution, figSize, showPlot)
    elif (solutionName == "numericalApproxBound"):
        appxsolution = solver.ua
        solution = solver.unab
        PlotUtils.plotDifference(domains, "Figures/5_Difference_in_Numerical_solution_approx_boundary", mesh, solver, appxsolution, solution, figSize, showPlot)
    elif (solutionName == "numericalIterBound"):
        appxsolution = solver.ua
        solution = solver.unib
        PlotUtils.plotDifference(domains, "Figures/5_Difference_in_Numerical_solution_iter_boundary", mesh, solver, appxsolution, solution, figSize, showPlot)

def showDSSDifference(mesh, solver, solutionName, figSize, showPlot):
    """
    shows difference in dss solution.

    Parameters
    ----------
    mesh : object
        mesh object
    solver : object
        solver object
    solutionName : str
        name of solution
    figSize : list
        size of figure
    showPlot : boolean
        show plot if true

    Returns
    -------
    None
    """

    domains = joblib.load("Data/initialDomains.sav")
    if (solutionName == "approximate"):
        print("Difference is not supported for approximate solution")
    elif (solutionName == "numerical"):
        appxsolution = solver.ua
        solution = solver.un
        PlotUtils.plotDifference(domains, "Figures/5_Difference_in_Numerical_DSS_solution", mesh, solver, appxsolution, solution, figSize, showPlot)
    elif (solutionName == "numericalApproxBound"):
        appxsolution = solver.ua
        solution = solver.unab
        PlotUtils.plotDifference(domains, "Figures/5_Difference_in_Numerical_DSS_solution_approx_boundary", mesh, solver, appxsolution, solution, figSize, showPlot)
    elif (solutionName == "numericalIterBound"):
        appxsolution = solver.ua
        solution = solver.unib
        PlotUtils.plotDifference(domains, "Figures/5_Difference_in_Numerical_DSS_solution_iter_boundary", mesh, solver, appxsolution, solution, figSize, showPlot)

def showDistributedSources(initialDomains, mesh, solver, solutionName, colorbarPosition, figSize, showPlot):
    """
    shows distributed sources.

    Parameters
    ----------
    initialDomains : list
        list of geometry objects
    mesh : object
        mesh object
    solver : object
        solver object
    solutionName : str
        name of solution
    colorbarPosition : str
        position of colorbar
    figSize : list
        size of figure
    showPlot : boolean
        show plot if true

    Returns
    -------
    None
    """
    
    if (solutionName == "approximate"):
        print("Difference not support approximate solution")
    else:
        PlotUtils.plotContour(initialDomains, mesh, solver, "SourceDSS", "Figures/6_Difference_in_Distributed_sources.png", colorbarPosition, figSize, showPlot)




import matplotlib.pyplot as plt
import numpy as np

from scipy.spatial import Delaunay as delaunay

class Mesh:
    """
    A class to represent mesh.
    
    ...

    Attributes
    ----------
    ix : list
        x positions of nodes
    iy : list
        y positions of nodes
    x : list
        x positions of nodes
    y : list
        y positions of nodes
    tri : list
        connectivity matrix
    

    Methods
    -------
    getNeiElements():
        generates list of neighbouring elements for each node
    getNeiNodes():
        generates list of neighbouring nodes for each node
    """

    def __init__(self, ix, iy, tri):
        """
        Constructs all the necessary attributes for the geometry object.

        Parameters
        ----------
        ix : list
            x positions of nodes
        iy : list
            y positions of nodes
        x : list
            x positions of nodes
        y : list
            y positions of nodes
        tri : list
            connectivity matrix
        """

        self.ix = ix
        self.iy = iy
        self.x = ix
        self.y = iy
        self.tri = tri

    def getNeiElements(self):
        """
        generates list of neighbouring elements for each node.

        Parameters
        ----------
        None

        Returns
        -------
        None
        """

        nn = len(self.ix)
        ne = len(self.tri[:,0])
        neiElements = []

        for i in range(nn):
            ele = []
    
            for j in range(ne):

                if self.tri[j,0] == i or self.tri[j,1] == i or self.tri[j,2] == i:
                    ele.append(j)

            neiElements.append(ele)

        self.neiElements = neiElements

    def getNeiNodes(self):
        """
        generates list of neighbouring nodes for each node.

        Parameters
        ----------
        None

        Returns
        -------
        None
        """

        nn = len(self.ix)
        neiNodes = []

        for i in range(nn):

            nod = []
            for j in self.neiElements[i]:

                for k in self.tri[j]:

                    if k != i: 

                        nod.append(k)

            nod = np.unique(np.array(nod))
            nod = nod.tolist()

            neiNodes.append(nod)

        self.neiNodes = neiNodes

def generateNodes(initialDomains, meshingMethod, figSize, showPlot):
    """
    gets nodes.

    Parameters
    ----------
    initialDomains : list
        list of geometries
    meshingMethod : str
        method of meshing
    figSize : list
        size of figure
    showPlot : boolean
        show plot if True

    Returns
    -------
    None
    """

    markers = ["+", "_", "*", "^", "x", "o"]
    plt.close("all")
    # Plotting the geometry again
    if figSize == "Default":
        figSize = initialDomains[0].figSize

    fig, ax = plt.subplots(figsize= figSize)
    w = initialDomains[0].width
    h = initialDomains[0].height
    [xp,yp] = initialDomains[0].position
    dx = w/100
    sf = 1.5
    plt.xlim(xp-w/2-dx, xp+w/2+dx)
    plt.ylim(yp-h/2-dx, yp+h/2+dx)

    lW = (figSize[0]/initialDomains[0].defaultFigSize[0])*0.5
    for domain in initialDomains:
        domain.show(plt, lW)

    # Generating additional data required for meshing

    # Relative width
    for domain in initialDomains:
        domain.calculateRelativeWidth(initialDomains[0].width)

    # Width Nodes
    for domain in initialDomains:
        domain.calculateWidthNodes(initialDomains[0].non)

    # Area
    for domain in initialDomains:
        domain.calculateArea()

    # Maximum number of free nodes
    for domain in initialDomains:
        domain.calculateMaxFreeNodes(initialDomains)

    # Actual meshing starts from here

    # Generate vertices
    count = 0
    for domain in initialDomains:
        domain.genVertices(plt, lW, markers[count % len(markers)])
        count = count + 1

    # Generate boundary nodes
    count = 0
    for domain in initialDomains:
        domain.genBoundaryNodes(plt, lW, markers[count % len(markers)])
        count = count + 1

    # Generate internal free nodes

    for domain in initialDomains:
        domain.isMeshed = False


    for j in range(len(initialDomains)): # Atleast one domain will be meshed in the next loop
                                        # This loop is to take care of the worst case scenario
        count = 0
        for domain in initialDomains: 
            
            if domain.isParent: # Dealing with parents with childs

                if (not domain.isMeshed): # Skip if already meshed

                    # Assume the childs are not meshed
                    domain.childMeshed = False
                    mc = 0 # Initialise no. of meshed childs
                    cd = domain.childDomains

                    # Check if all child domains are meshed
                    for k in cd:
                        if initialDomains[k].isMeshed:
                            mc = mc+1

                    if mc == len(cd):
                        domain.childMeshed = True

                    # Start meshing if all childs are meshed
                    if domain.childMeshed:
                        print("Meshing domain number", count)
                        if meshingMethod == "tri":
                            domain.genTriFreeNodes(initialDomains, plt, lW, markers[count % len(markers)])
                        elif meshingMethod == "rand":
                            domain.genRandFreeNodes(initialDomains, plt, lW, markers[count % len(markers)])
                        domain.isMeshed = True
                        
            else: # Dealing with pure child domains (Pure childs will be meshed first)

                if (not domain.isMeshed): # Skip if already meshed

                    print("Meshing domain number", count)
                    if meshingMethod == "tri":
                        domain.genTriFreeNodes(initialDomains, plt, lW, markers[count % len(markers)])
                    elif meshingMethod == "rand":
                        domain.genRandFreeNodes(initialDomains, plt, lW, markers[count % len(markers)])
                    domain.isMeshed = True

            count = count +1

    plt.gca().set_aspect("equal")
    for axis in ['top', 'bottom', 'left', 'right']:
        ax.spines[axis].set_linewidth(figSize[0]/initialDomains[0].defaultFigSize[0])
    
    plt.xlabel("x - axis [meter]", fontsize=figSize[0]/initialDomains[0].defaultFigSize[0]*1.5*20*sf)
    plt.ylabel("y - axis [meter]", fontsize=figSize[0]/initialDomains[0].defaultFigSize[0]*1.5*20*sf)
    plt.xticks(fontsize= figSize[0]/initialDomains[0].defaultFigSize[0]*20*sf)
    plt.yticks(fontsize= figSize[0]/initialDomains[0].defaultFigSize[0]*20*sf)
    plt.grid()
    plt.tick_params(axis='both', width=figSize[0]/initialDomains[0].defaultFigSize[0])
    plt.tick_params(axis='both', length=figSize[0]/initialDomains[0].defaultFigSize[0]*5)
    plt.savefig("Figures/1_Nodes.png")
    if showPlot == True:
        plt.show()

def createMesh(initialDomains):
    """
    gets mesh.

    Parameters
    ----------
    initialDomains : list
        list of geometries
    
    Returns
    -------
    None
    """

    # initialDomains = joblib.load("Data/initialDomains.sav")
    print("Creating mesh")
    # Generate boundary, free node list, x, y positions of all nodes

    x = np.array([])
    y = np.array([])
    nn = -1
    for domain in initialDomains:

        if domain.shape == "point":
            domain.boundaryList = np.array([nn+1])
            nn = nn+1
            domain.freeList = np.array([])
            x =  np.concatenate((x, [domain.boundaryNodesX]))
            y =  np.concatenate((y, [domain.boundaryNodesY]))

        elif domain.shape == "line":
            domain.boundaryList = np.arange(nn+1, nn+len(domain.boundaryNodesX)+1, 1)
            nn = domain.boundaryList[-1]
            domain.freeList = np.array([])
            x =  np.concatenate((x, domain.boundaryNodesX))
            y =  np.concatenate((y, domain.boundaryNodesY))

        else:
            domain.boundaryList = np.arange(nn+1, nn+len(domain.boundaryNodesX)+1, 1)
            nn = domain.boundaryList[-1]
            if len(domain.freeNodesX) == 0:
                domain.freeList = np.array([])
            else:
                domain.freeList = np.arange(nn+1, nn+len(domain.freeNodesX)+1, 1)
                nn = domain.freeList[-1]

            x =  np.concatenate((x, domain.boundaryNodesX, domain.freeNodesX))
            y =  np.concatenate((y, domain.boundaryNodesY, domain.freeNodesY))

    xy = np.transpose([x, y])
    tri = delaunay(xy)
    mesh = Mesh(x, y, tri.simplices)

    # Generate element list in each domain

    ne = len(tri.simplices[:,0])
    nn = len(mesh.ix)

    for domain in initialDomains:
        elementList = [] # Initializing element list

        if domain.shape != "point" and domain.shape != "line":

            if domain.parentDomains == []: # Checking for global domain to also include boundary nodes for element extraction
                if len(domain.freeNodesX) == 0:
                    domainNodes = domain.boundaryList
                else:
                    domainNodes = np.concatenate(domain.boundaryList, domain.freeList)
            else:
                domainNodes = domain.freeList
                boundaryNodes = domain.boundaryList
                
            domainNodes = domainNodes.astype(int)

            for cd in domain.childDomains:
                if initialDomains[cd].shape == "line":
                    domainNodes = np.concatenate((domainNodes, initialDomains[cd].boundaryList))

            for j in range(ne):

                for k in domainNodes:

                    if tri.simplices[j,0] == k or tri.simplices[j,1] == k or tri.simplices[j,2] == k:
                        elementList.append(j)
                        break

        domain.elementList = elementList

    # To assign interior elements formed by only the boundary nodes

    actualElementList = np.arange(0,ne,1).tolist()
    currentElementList = np.array([]).tolist()
    for domain in initialDomains:
        currentElementList = currentElementList + domain.elementList

    remainingElements = actualElementList.copy()

    for element in actualElementList:
        if element in currentElementList:
            remainingElements.remove(element)

    for domain in initialDomains:
        boundaryNodes = domain.boundaryList
        for element in remainingElements:
            for node in boundaryNodes:
                if node in mesh.tri[element]:
                    domain.elementList = domain.elementList + [element]
                    break

    # Extracting neighbour information
    print("Extracting neighbour information")
    # Elements
    mesh.getNeiElements()
    # Nodes
    mesh.getNeiNodes()
    return mesh

def relaxMesh(initialDomains, mesh):
    """
    relaxes the mesh.

    Parameters
    ----------
    initialDomains : list
        list of geometries
    mesh : object
        the mesh object

    Returns
    -------
    None
    """

    print("Relaxing mesh")
    
    rx = mesh.ix
    ry = mesh.iy
    nn = len(rx)

    Fx = np.linspace(0, 0, len(rx))
    Fy = Fx

    rMin = initialDomains[0].width/initialDomains[0].non

    for rm in range(1000):

        print("Percentage of relaxation", rm/1000*100, end='\r')

        for domain in initialDomains:

            if domain.shape != "point" and domain.shape != "line":

                frn = domain.freeList

                for j in frn:
                    nein = mesh.neiNodes[j]
                    Fx[j] = 0
                    Fy[j] = 0

                    for k in nein:
                        stiff = 1
                        rmag = np.sqrt((rx[k]-rx[j])**2+(ry[k]-ry[j])**2)
                        Fx[j] = Fx[j] + stiff*(rx[k]-rx[j])/rmag**0.1
                        Fy[j] = Fy[j] + stiff*(ry[k]-ry[j])/rmag**0.1

                for j in frn:
                    if Fx[j] != 0 and Fy[j] != 0:
                        rx[j] = rx[j] + 0.002*rMin*Fx[j]/np.sqrt(Fx[j]**2+Fy[j]**2)
                        ry[j] = ry[j] + 0.002*rMin*Fy[j]/np.sqrt(Fx[j]**2+Fy[j]**2)

    print()

    mesh.rx = rx
    mesh.ry = ry
    mesh.x = rx
    mesh.y = ry

def assignSources(initialDomains, mesh, solver):
    """
    assigns sources.

    Parameters
    ----------
    initialDomains : list
        list of geometries
    mesh : object
        the mesh object
    solver : object
        the solver object

    Returns
    -------
    None
    """

    x = mesh.rx
    y = mesh.ry

    nn = len(x)
    # nodeSource is for both plotting and matrix assignment
    # elementSource may serve some purpose in different physics
    if solver.name == 'structural':

        nodeSource = np.zeros((nn,2))
        ne = len(mesh.tri[:,0])
        elementSource = np.zeros((ne,2))
        print("Assigning sources")

        for domain in initialDomains:

            node = domain.sourcePosition
            nodeSource[node] = domain.source

    else:
        nodeSource = np.zeros((nn,1))
        ne = len(mesh.tri[:,0])
        elementSource = np.zeros((ne,1))
        print("Assigning sources")

        for domain in initialDomains:

            if domain.shape == "point" or domain.shape == "line":
                nodes = domain.boundaryList
                for node in nodes:
                    nodeSource[node] = domain.source/len(nodes)
                    elements = mesh.neiElements[node]
                    for element in elements:
                        elementSource[element] = elementSource[element] + domain.source/len(elements)

            else:
                elements = domain.elementList
                for element in elements:
                    ex = []
                    ey = []
                    for node in mesh.tri[element]:
                        ex.append(x[node])
                        ey.append(y[node])

                    p = [ey[1]-ey[2], ey[2]-ey[0], ey[0]-ey[1]]
                    q = [ex[2]-ex[1], ex[0]-ex[2], ex[1]-ex[0]]
                    ea = 0.5*abs(p[1]*q[2]-q[1]*p[2])

                    if ea == 0:
                        print("element area is zero at: ", element)

                    elementSource[element] = domain.source*ea

                for element in elements:

                    for node in mesh.tri[element]:
                        nodeSource[node] = nodeSource[node] + elementSource[element]/3

        mesh.elementSource = elementSource
                    
    mesh.nodeSource = nodeSource
    
def assignMaterials(initialDomains, mesh):
    """
    assigns materials.

    Parameters
    ----------
    initialDomains : list
        list of geometries
    mesh : object
        the mesh object

    Returns
    -------
    None
    """

    # nodeMaterial is for plotting
    # elementMaterial is for matrix computation

    x = mesh.rx
    nn = len(x)
    nodeMaterial = np.ones((nn,1))
    ne = len(mesh.tri[:,0])
    elementMaterial = np.ones((ne,1))
    print("Assigning material properties")

    for domain in initialDomains:

        if domain.shape == "point" or domain.shape == "line":
            nodes = domain.boundaryList
            for node in nodes:
                nodeMaterial[node] = domain.material

        else:
            nodes = np.concatenate((domain.boundaryList, domain.freeList)).astype(int)
            for node in nodes:
                nodeMaterial[node] = domain.material

        elements = domain.elementList
        for element in elements:
            elementMaterial[element] = domain.material

    mesh.nodeMaterial = nodeMaterial
    mesh.elementMaterial = elementMaterial

def getNeighbourInformation(mesh, maxTn):
    """
    gets neighbour information for DSS.

    Parameters
    ----------
    mesh : object
        the mesh object
    maxTn : int
        truncation number, all the lower numbered information will be automatically generated

    Returns
    -------
    None
    """
        
    tri = mesh.tri
    ne = len(tri[:,0])
    nn = len(mesh.rx)

    print("Generating neighbour element and node list upto Truncation Number", maxTn)
    eleAttNames = []
    nodeAttNames = []
    farnodeAttNames = []

    for i in range(maxTn+1): # Initializing all truncated element list
        globals()['neiElementsTn%s' % i] = np.empty([nn,0]).tolist()
        setattr(mesh, 'neiElementsTn%s' % i, globals()['neiElementsTn%s' % i])
        eleAttNames.append('neiElementsTn%s' % i)

    for i in range(nn):
        for j in range(ne):
            if tri[j,0] == i or tri[j,1] == i or tri[j,2] == i:
                mesh.neiElementsTn0[i].append(j)

    for i in range(maxTn+1): # Initializing all truncated node list
        globals()['neiNodesTn%s' % i] = np.empty([nn,0]).tolist()
        setattr(mesh, 'neiNodesTn%s' % i, globals()['neiNodesTn%s' % i])
        nodeAttNames.append('neiNodesTn%s' % i)

    for i in range(nn):
        mesh.neiNodesTn0[i].append(i) # Redundant. However, completes the set.

    for i in range(1, maxTn+1): # Initializing all far neighbour list
        globals()['farneiNodesTn%s' % i] = np.empty([nn,0]).tolist()
        setattr(mesh, 'farneiNodesTn%s' % i, globals()['farneiNodesTn%s' % i])
        farnodeAttNames.append('farneiNodesTn%s' % i)

    for tn in range(1,maxTn+1):
        neiNodesTn = np.empty([nn,0]).tolist()
        neiElementsTn = np.empty([nn,0]).tolist()
        farneiNodesTn = np.empty([nn,0]).tolist()

        for i in range(nn):
            existingNodes = getattr(mesh, nodeAttNames[tn-1])[i]
            newNodes = []

            for node in existingNodes:
                potElements = getattr(mesh, eleAttNames[0])[node]
                for k in potElements:
                    for l in range(3):
                        newNodes.append(tri[k,l])

            newNodes = np.unique(newNodes)
            newNodesList = newNodes.tolist()

            for enode in existingNodes:
                if enode in newNodes:
                    newNodesList.remove(enode)

            newNodes = newNodesList

            totalNodes = existingNodes.copy()
            for node in newNodes:
                totalNodes.append(node)

            neiNodesTn[i] = totalNodes.copy()
            farneiNodesTn[i] = newNodes.copy()

            truncatedElements = []

            for node in existingNodes:
                potElements = getattr(mesh, eleAttNames[0])[node]
                for k in potElements:
                    truncatedElements.append(k)

            truncatedElements = np.unique(truncatedElements)

            neiElementsTn[i] = truncatedElements.tolist()

        print("Current truncation number", tn, end='\r')
        setattr(mesh, nodeAttNames[tn], neiNodesTn)
        setattr(mesh, eleAttNames[tn], neiElementsTn)
        setattr(mesh, farnodeAttNames[tn-1], farneiNodesTn)

    print(end= '\r')
    mesh.eleAttNames = eleAttNames
    mesh.nodeAttNames = nodeAttNames
    mesh.farnodeAttNames = farnodeAttNames

            
            







import numpy as np
from scipy.linalg import solve

class Solver():
    """
    A class to represent a numerical solver.
    
    ...

    Attributes
    ----------
    None

    Methods
    -------
    solveApproximateMatricesSadiku(A, ui, b, target):
        solves given matrices
    solveApproximateMatricesScipy(A, b):
        solves given matrices
    solveApproximateMatricesNumpy(A, b):
        solves given matrices
    solveNumericalMatricesSadiku(A, ui, b, target):
        solves given matrices
    solveNumericalMatricesScipy(A, b):
        solves given matrices
    solveNumericalMatricesNumpy(A, b):
        solves given matrices
    solveApproximateBoundaryMatricesSadiku(A, ui, b, target):
        solves given matrices
    solveApproximateBoundaryMatricesScipy(A, b):
        solves given matrices
    solveApproximateBoundaryMatricesNumpy(A, b):
        solves given matrices
    solveNumericalMatricesApproxBoundSadiku(A, ui, b, target):
        solves given matrices
    solveNumericalMatricesApproxBoundScipy(A, b):
        solves given matrices
    solveNumericalMatricesApproxBoundNumpy(A, b):
        solves given matrices
    solveNumericalMatricesApproxIterBoundSadiku(A, ui, b, target):
        solves given matrices
    solveNumericalMatricesApproxIterBoundScipy(A, b):
        solves given matrices
    solveNumericalMatricesApproxIterBoundNumpy(A, b):
        solves given matrices
    solveIterBoundaryMatricesSadiku(A, ui, b, target):
        solves given matrices
    solveIterBoundaryMatricesScipy(A, b):
        solves given matrices
    solveIterBoundaryMatricesNumpy(A, b):
        solves given matrices
    solveminiScipy(A, b):
        solves given matrices
    solveminiNumpy(A, b):
        solves given matrices
    solveminiSadiku(A, ui, b, target):
        solves given matrices
    """
    
    def __init__(self):
        """
        Constructs all the necessary attributes for the Solver object.

        Parameters
        ----------
        None
        """

        pass
        

    def solveApproximateMatricesSadiku(self, A, ui, b, target):
        """
        calculates solution of system of equations represented in matrix form.

        Parameters
        ----------
        A : list
            coeffecient matrix
        ui : list
            initial guess
        b : list
            forcing matrix
        target : float
            target tolerence
        Returns
        -------
        None
        """

        u = ui.copy()
        error = 1

        while (error > target):
            up = u.copy()
            for i in range(len(u)):
                Az = A[i,:].copy()
                Az[i] = 0
                u[i] = (np.matmul(-Az,up)+b[i])/A[i,i]
                
            error = np.amax(np.absolute((u-up)/up*100))

        self.ua = u
        self.u = u

    def solveApproximateMatricesScipy(self, A, b):
        """
        calculates solution of system of equations represented in matrix form.

        Parameters
        ----------
        A : list
            coeffecient matrix
        b : list
            forcing matrix
        Returns
        -------
        None
        """

        self.ua = solve(A, b)
        self.u = self.ua.copy()

    def solveApproximateMatricesNumpy(self, A, b):
        """
        calculates solution of system of equations represented in matrix form.

        Parameters
        ----------
        A : list
            coeffecient matrix
        b : list
            forcing matrix
        Returns
        -------
        None
        """

        self.ua = np.linalg.lstsq(A, b)[0]
        self.u = self.ua.copy()

    def solveNumericalMatricesSadiku(self, A, ui, b, target):
        """
        calculates solution of system of equations represented in matrix form.

        Parameters
        ----------
        A : list
            coeffecient matrix
        ui : list
            initial guess
        b : list
            forcing matrix
        target : float
            target tolerence
        Returns
        -------
        None
        """

        u = ui.copy()
        error = 1

        while (error > target):
            up = u.copy()
            for i in range(len(u)):
                Az = A[i,:].copy()
                Az[i] = 0
                u[i] = (np.matmul(-Az,up)+b[i])/A[i,i]
                
            error = np.amax(np.absolute((u-up)/up*100))

        self.un = u
        self.u = u

    def solveNumericalMatricesScipy(self, A, b):
        """
        calculates solution of system of equations represented in matrix form.

        Parameters
        ----------
        A : list
            coeffecient matrix
        b : list
            forcing matrix
        Returns
        -------
        None
        """

        self.un = solve(A, b)
        self.u = self.un.copy()

    def solveNumericalMatricesNumpy(self, A, b):
        """
        calculates solution of system of equations represented in matrix form.

        Parameters
        ----------
        A : list
            coeffecient matrix
        b : list
            forcing matrix
        Returns
        -------
        None
        """

        self.un = np.linalg.lstsq(A, b)[0]
        self.u = self.un.copy()

    def solveApproximateBoundaryMatricesSadiku(self, A, ui, b, target):
        """
        calculates solution of system of equations represented in matrix form.

        Parameters
        ----------
        A : list
            coeffecient matrix
        ui : list
            initial guess
        b : list
            forcing matrix
        target : float
            target tolerence
        Returns
        -------
        None
        """

        u = ui.copy()
        error = 1

        while (error > target):
            up = u.copy()
            for i in range(len(u)):
                Az = A[i,:].copy()
                Az[i] = 0
                u[i] = (np.matmul(-Az,up)+b[i])/A[i,i]
                
            error = np.amax(np.absolute((u-up)/up*100))

        self.uab = u
        self.u = u

    def solveApproximateBoundaryMatricesScipy(self, A, b):
        """
        calculates solution of system of equations represented in matrix form.

        Parameters
        ----------
        A : list
            coeffecient matrix
        b : list
            forcing matrix
        Returns
        -------
        None
        """

        self.uab = solve(A, b)
        self.u = self.uab.copy()

    def solveApproximateBoundaryMatricesNumpy(self, A, b):
        """
        calculates solution of system of equations represented in matrix form.

        Parameters
        ----------
        A : list
            coeffecient matrix
        b : list
            forcing matrix
        Returns
        -------
        None
        """

        self.uab = np.linalg.lstsq(A, b)[0]
        self.u = self.uab.copy()

    def solveNumericalMatricesApproxBoundSadiku(self, A, ui, b, target):
        """
        calculates solution of system of equations represented in matrix form.

        Parameters
        ----------
        A : list
            coeffecient matrix
        ui : list
            initial guess
        b : list
            forcing matrix
        target : float
            target tolerence
        Returns
        -------
        None
        """

        u = ui.copy()
        error = 1

        while (error > target):
            up = u.copy()
            for i in range(len(u)):
                Az = A[i,:].copy()
                Az[i] = 0
                u[i] = (np.matmul(-Az,up)+b[i])/A[i,i]
                
            error = np.amax(np.absolute((u-up)/up*100))

        self.unab = u
        self.u = u

    def solveNumericalMatricesApproxBoundScipy(self, A, b):
        """
        calculates solution of system of equations represented in matrix form.

        Parameters
        ----------
        A : list
            coeffecient matrix
        b : list
            forcing matrix
        Returns
        -------
        None
        """

        self.unab = solve(A, b)
        self.u = self.unab.copy()

    def solveNumericalMatricesApproxBoundNumpy(self, A, b):
        """
        calculates solution of system of equations represented in matrix form.

        Parameters
        ----------
        A : list
            coeffecient matrix
        b : list
            forcing matrix
        Returns
        -------
        None
        """

        self.unab = np.linalg.lstsq(A, b)[0]
        self.u = self.unab.copy()

    def solveNumericalMatricesApproxIterBoundSadiku(self, A, ui, b, target):
        """
        calculates solution of system of equations represented in matrix form.

        Parameters
        ----------
        A : list
            coeffecient matrix
        ui : list
            initial guess
        b : list
            forcing matrix
        target : float
            target tolerence
        Returns
        -------
        None
        """

        u = ui.copy()
        error = 1

        while (error > target):
            up = u.copy()
            for i in range(len(u)):
                Az = A[i,:].copy()
                Az[i] = 0
                u[i] = (np.matmul(-Az,up)+b[i])/A[i,i]
                
            error = np.amax(np.absolute((u-up)/up*100))

        self.unib = u
        self.u = u

    def solveNumericalMatricesApproxIterBoundScipy(self, A, b):
        """
        calculates solution of system of equations represented in matrix form.

        Parameters
        ----------
        A : list
            coeffecient matrix
        b : list
            forcing matrix
        Returns
        -------
        None
        """

        self.unib = solve(A, b)
        self.u = self.unib.copy()

    def solveNumericalMatricesApproxIterBoundNumpy(self, A, b):
        """
        calculates solution of system of equations represented in matrix form.

        Parameters
        ----------
        A : list
            coeffecient matrix
        b : list
            forcing matrix
        Returns
        -------
        None
        """

        self.unib = np.linalg.lstsq(A, b)[0]
        self.u = self.unib.copy()

    def solveIterBoundaryMatricesSadiku(self, A, ui, b, target):
        """
        calculates solution of system of equations represented in matrix form.

        Parameters
        ----------
        A : list
            coeffecient matrix
        ui : list
            initial guess
        b : list
            forcing matrix
        target : float
            target tolerence
        Returns
        -------
        None
        """

        u = ui.copy()
        error = 1

        while (error > target):
            up = u.copy()
            for i in range(len(u)):
                Az = A[i,:].copy()
                Az[i] = 0
                u[i] = (np.matmul(-Az,up)+b[i])/A[i,i]
                
            error = np.amax(np.absolute((u-up)/up*100))

        self.uib = u
        self.u = u

    def solveIterBoundaryMatricesScipy(self, A, b):
        """
        calculates solution of system of equations represented in matrix form.

        Parameters
        ----------
        A : list
            coeffecient matrix
        b : list
            forcing matrix
        Returns
        -------
        None
        """

        self.uib = solve(A, b)
        self.u = self.uib.copy()

    def solveIterBoundaryMatricesNumpy(self, A, b):
        """
        calculates solution of system of equations represented in matrix form.

        Parameters
        ----------
        A : list
            coeffecient matrix
        b : list
            forcing matrix
        Returns
        -------
        None
        """

        self.uib = np.linalg.lstsq(A, b)[0]
        self.u = self.uib.copy()
    
    def solveminiScipy(self, A, b):
        """
        calculates solution of system of equations represented in matrix form.

        Parameters
        ----------
        A : list
            coeffecient matrix
        b : list
            forcing matrix
        Returns
        -------
        None
        """

        self.miniu = solve(A, b)

    def solveminiNumpy(self, A, b):
        """
        calculates solution of system of equations represented in matrix form.

        Parameters
        ----------
        A : list
            coeffecient matrix
        b : list
            forcing matrix
        Returns
        -------
        None
        """
        
        self.miniu = np.linalg.lstsq(A, b)[0]
        
    def solveminiSadiku(self, A, ui, b, target):
        """
        calculates solution of system of equations represented in matrix form.

        Parameters
        ----------
        A : list
            coeffecient matrix
        ui : list
            initial guess
        b : list
            forcing matrix
        target : float
            target tolerence
        Returns
        -------
        None
        """

        u = ui.copy()
        error = 1

        while (error > target):
            up = u.copy()
            for i in range(len(u)):
                Az = A[i,:].copy()
                Az[i] = 0
                u[i] = (np.matmul(-Az,up)+b[i])/A[i,i]
                
            error = np.amax(np.absolute((u-up)/up*100))

        self.miniu = u


# Import packages
import joblib
import os
from DssPyLib.Domains.Rectangle import Rectangle
from DssPyLib.Domains.Ellipse import Ellipse
from DssPyLib.Domains.Line import Line
from DssPyLib.Domains.Point import Point

import DssPyLib.PlotUtils as PlotUtils
import Output

def Geometry(figSize = "Default", showPlot = False):
    """
    user input geometry.

    Parameters
    ----------
    figSize : list
        size of figure
    showPlot : boolean
        show plot if true

    Returns
    -------
    None
    """

    initialDomains = []
    # The first domain with index 0 is always a Global domain and should be initiated with non

    # Domain 0
    initialDomains.append(Rectangle(2, # Width
                                    1, # Height
                                    [0, 0], # Position
                                    [1, 1], # Relative Mesh Density
                                    0, # Angle
                                    False, # is_child
                                    True, # is_parent
                                    [2], # Child domains
                                    [], # Parent domains
                                    20)) # non, Predefined no. of nodes

    # Domain 1
    initialDomains.append(Point(0, # Width
                                0, # Height
                                [0, 0], # Position
                                [1, 1], # Relative Mesh Density
                                0, # Angle
                                True, # is_child
                                False, # is_parent
                                [], # Child domains
                                [2], # Parent domains
                                0)) # Predefined no. of nodes

    # Domain 2
    h = initialDomains[0].width/initialDomains[0].non
    initialDomains.append(Rectangle(initialDomains[0].width - 2*h, # Width
                                initialDomains[0].height - 2*h, # Height
                                initialDomains[0].position, # Position
                                [1.2, 1.2], # Relative Mesh Density
                                0, # Angle
                                True, # is_child
                                True, # is_parent
                                [1], # Child domains
                                [0], # Parent domains
                                0)) # Predefined no. of nodes

    if figSize == "Default":
        figSize = (20, 20 * initialDomains[0].height/ initialDomains[0].width)
    
    initialDomains[0].defaultFigSize = (20, 20 * initialDomains[0].height/ initialDomains[0].width)
    initialDomains[0].figSize = figSize
    print("The shape of global domain is", initialDomains[0].shape, "(From input.py)")
    
    if not os.path.exists('Data'):
        os.mkdir('Data')

    if not os.path.exists('Figures'):
        os.mkdir('Figures')

    joblib.dump(initialDomains, "Data/initialDomains.sav")
    PlotUtils.plotGeometry(initialDomains, "Figures/1_Geometry.png", figSize, showPlot)

def Sources(colorbarPosition = "right", figSize = "Default", showPlot = False):
    """
    user input sources.

    Parameters
    ----------
    colorbarPosition : str
        position of color bar
    figSize : list
        size of figure
    showPlot : boolean
        show plot if true

    Returns
    -------
    None
    """
    
    initialDomains = joblib.load("Data/initialDomains.sav")

    initialDomains[0].source = 0
    initialDomains[1].source = 1
    initialDomains[2].source = 0

    joblib.dump(initialDomains, "Data/initialDomains.sav")
    Output.assigedSources(colorbarPosition, figSize, showPlot)

def Materials(colorbarPosition = "right", figSize = "Default", showPlot = False):
    """
    user input materials.

    Parameters
    ----------
    colorbarPosition : str
        position of color bar
    figSize : list
        size of figure
    showPlot : boolean
        show plot if true

    Returns
    -------
    None
    """
    
    # Material property of points and lines will not be considered as material property is bulk property 
    initialDomains = joblib.load("Data/initialDomains.sav")

    initialDomains[0].material = 1
    initialDomains[1].material = 1
    initialDomains[2].material = 1

    joblib.dump(initialDomains, "Data/initialDomains.sav")
    Output.assigedMaterials(colorbarPosition, figSize, showPlot)

def Solver(solverName):
    """
    user input solver.

    Parameters
    ----------
    solverName : str
        name of the solver

    Returns
    -------
    None
    """

    if (solverName == "electrostatic"):
        from DssPyLib.Solvers.Electrostatic import Electrostatic
        solver = Electrostatic()
    elif (solverName == "magnetostatic"):
        from DssPyLib.Solvers.Magnetostatic import Magnetostatic
        solver = Magnetostatic()
    joblib.dump(solver, "Data/solver.sav")

if __name__ == "__main__":
    Geometry()


import Input
import Output
import Get
import numpy as np

def main():
    Input.Geometry(figSize = "Default", showPlot = False)

    Output.nodes(meshingMethod = "rand", figSize = "Default", showPlot = False)
    Output.mesh(meshNumbering = "numbered", figSize = "Default", showPlot = False)
    Output.relaxedMesh(meshNumbering = "numbered", figSize = "Default", showPlot = False)

    Input.Solver("electrostatic")
    Input.Sources(colorbarPosition = "right", figSize = "Default", showPlot = False)
    Input.Materials(colorbarPosition = "right", figSize = "Default", showPlot = False)

    Output.neighbourInformation(truncationNumber = 6)
    Output.solution(solutionName = "approximate", solverMethod = "Scipy", figSize = "Default", colorbarPosition = "right", showPlot = False, withField = False, withDifference = True)

    tht = range(0, 360, 20)
    rad = 0.3
    ep0 = 8.8541878128*10**(-12)

    E_magnitude_analytical = 1/(2*np.pi*ep0*rad)
    print("Magnitude of Analytical Electric FIeld is", E_magnitude_analytical)
    for th in tht:
        x = rad*np.cos(np.deg2rad(th))
        y = rad*np.sin(np.deg2rad(th))
        position = np.array([[x, y]])
        field = Get.vectorField(position)
        print(field, (np.linalg.norm(field)-E_magnitude_analytical)/E_magnitude_analytical*100)

    Output.solution(solutionName = "numerical", solverMethod = "Scipy", figSize = "Default", colorbarPosition = "right", showPlot = False, withField = False, withDifference = True)

    for th in tht:
        x = rad*np.cos(np.deg2rad(th))
        y = rad*np.sin(np.deg2rad(th))
        position = np.array([[x, y]])
        field = Get.vectorField(position)
        print(field, (np.linalg.norm(field)-E_magnitude_analytical)/E_magnitude_analytical*100)

    Output.solution(solutionName = "numericalApproxBound", solverMethod = "Scipy", figSize = "Default", colorbarPosition = "right", showPlot = False, withField = False, withDifference = True)

    for th in tht:
        x = rad*np.cos(np.deg2rad(th))
        y = rad*np.sin(np.deg2rad(th))
        position = np.array([[x, y]])
        field = Get.vectorField(position)
        print(field, (np.linalg.norm(field)-E_magnitude_analytical)/E_magnitude_analytical*100)

    Output.solution(solutionName = "numericalIterBound", solverMethod = "Scipy", figSize = "Default", colorbarPosition = "right", showPlot = False, withField = False, withDifference = True)

    for th in tht:
        x = rad*np.cos(np.deg2rad(th))
        y = rad*np.sin(np.deg2rad(th))
        position = np.array([[x, y]])
        field = Get.vectorField(position)
        print(field, (np.linalg.norm(field)-E_magnitude_analytical)/E_magnitude_analytical*100)

    Output.dssSolution(solutionName = "numerical", truncationNumber = 5, solverMethod = "Scipy", figSize = "Default", colorbarPosition = "right", showPlot = False, withField = True, withDifference = True)

    for th in tht:
        x = rad*np.cos(np.deg2rad(th))
        y = rad*np.sin(np.deg2rad(th))
        position = np.array([[x, y]])
        field = Get.vectorField(position)
        print(field, (np.linalg.norm(field)-E_magnitude_analytical)/E_magnitude_analytical*100)

    Output.dssSolution(solutionName = "numericalApproxBound", truncationNumber = 5, solverMethod = "Scipy", figSize = "Default", colorbarPosition = "right", showPlot = False, withField = False, withDifference = True)

    for th in tht:
        x = rad*np.cos(np.deg2rad(th))
        y = rad*np.sin(np.deg2rad(th))
        position = np.array([[x, y]])
        field = Get.vectorField(position)
        print(field, (np.linalg.norm(field)-E_magnitude_analytical)/E_magnitude_analytical*100)

    Output.dssSolution(solutionName = "numericalIterBound", truncationNumber = 5, solverMethod = "Scipy", figSize = "Default", colorbarPosition = "right", showPlot = False, withField = False, withDifference = True)

    for th in tht:
        x = rad*np.cos(np.deg2rad(th))
        y = rad*np.sin(np.deg2rad(th))
        position = np.array([[x, y]])
        field = Get.vectorField(position)
        print(field, (np.linalg.norm(field)-E_magnitude_analytical)/E_magnitude_analytical*100)


if __name__ == "__main__":
    main()


import numpy as np

from . import Domain

class Point(Domain.Domain):
    def __init__(self, width, height, position, relativeMeshDensity, angle, isChild, isParent, childDomains, parentDomains, non):
        """
        Constructs all the necessary attributes for the geometry object.

        Parameters
        ----------
        shape: str
            the name of the shape of geometry
        """
        
        Domain.Domain.__init__(self, width, height, position, relativeMeshDensity, angle, isChild, isParent, childDomains, parentDomains, non)
        self.shape = "point"

    def show(self, plt, lW):
        """
        adds the current geometry to plot

        Parameters
        ----------
        plt : object
            matplotlib object
        lw : float
            width of line stroke

        Returns
        -------
        None
        """

        # Position of the point
        pos = np.array(self.position)
        # Plot the point
        plt.plot(pos[0], pos[1], 'k.', ms = lW*10)

    def calculateRelativeWidth(self, globalDomain):
        self.relativeWidth = 0

    def calculateArea(self):
        """
        calculates area.

        Parameters
        ----------
        None

        Returns
        -------
        None
        """

        self.area = 0

    def calculateMaxFreeNodes(self, domains):
        """
        calculates maximum number of free nodes allowed

        Parameters
        ----------
        domains : list
            list of all geometry objects

        Returns
        -------
        None
        """

        self.maxFreeNodes = 0

    def genVertices(self, plt, lW, markerShape):
        """
        gets vertices of the geometry.

        Parameters
        ----------
        plt : object
            matplotlib object
        lw : float
            width of line stroke
        markerShape : char
            shape of marker

        Returns
        -------
        None
        """

        # Generating vertices
        self.verticesX = np.array(self.position[0])
        self.verticesY = np.array(self.position[1])
        # Plot the vertices
        plt.scatter(self.verticesX, self.verticesY, c=None, marker=markerShape, s = lW*1000)

    def genBoundaryNodes(self, plt, lW, markerShape):
        """
        gets boundary nodes.

        Parameters
        ----------
        plt : object
            matplotlib object
        lw : float
            width of line stroke
        markerShape : char
            shape of marker

        Returns
        -------
        None
        """

        self.boundaryNodesX = self.verticesX
        self.boundaryNodesY = self.verticesY
        plt.scatter(self.boundaryNodesX, self.boundaryNodesY, c=None, marker=markerShape, s = lW*1000)

    def genRandFreeNodes(self, initialDomains, plt, lW, markerShape):
        """
        gets random free nodes inside a domain.

        Parameters
        ----------
        initialDomains : list
            list of all geometry objects
        plt : object
            matplotlib object
        lw : float
            width of line stroke
        markerShape : char
            shape of marker

        Returns
        -------
        None
        """

        self.freeNodesX = np.array([])
        self.freeNodesY = np.array([])
        self.allNodesX = self.boundaryNodesX
        self.allNodesY = self.boundaryNodesY

    def genTriFreeNodes(self, initialDomains, plt, lW, markerShape):
        """
        gets uniformly distributed triangular free nodes inside a domain.

        Parameters
        ----------
        initialDomains : list
            list of all geometry objects
        plt : object
            matplotlib object
        lw : float
            width of line stroke
        markerShape : char
            shape of marker

        Returns
        -------
        None
        """
        
        self.freeNodesX = np.array([])
        self.freeNodesY = np.array([])
        self.allNodesX = self.boundaryNodesX
        self.allNodesY = self.boundaryNodesY



from matplotlib import rcParams
rcParams['font.family'] = 'Times New Roman'
import matplotlib.pyplot as plt

import matplotlib.tri as mtri
import numpy as np
from matplotlib.ticker import FormatStrFormatter

def plotGeometry(domains, filename, figSize, showPlot):
    """
    plots geometry.

    Parameters
    ----------
    domains : list
        list of geometries
    filename : str
        name of file
    figSize : list
        size of figure
    showPlot : boolean
        show plot if True

    Returns
    -------
    None
    """

    plt.close("all")
    fig, ax = plt.subplots(figsize=domains[0].figSize)
    w = domains[0].width
    h = domains[0].height
    [xp,yp] = domains[0].position
    dx = w/100
    sf = 1.5
    plt.xlim(xp-w/2-dx, xp+w/2+dx)
    plt.ylim(yp-h/2-dx, yp+h/2+dx)

    lW = figSize[0]/domains[0].defaultFigSize[0]
    for domain in domains:
        domain.show(plt, lW)

    plt.gca().set_aspect("equal")
    for axis in ['top', 'bottom', 'left', 'right']:
        ax.spines[axis].set_linewidth(figSize[0]/domains[0].defaultFigSize[0])

    plt.xlabel("x - axis [meter]", fontsize=domains[0].figSize[0]*1.5*sf)
    plt.ylabel("y - axis [meter]", fontsize=domains[0].figSize[0]*1.5*sf)
    plt.xticks(fontsize= domains[0].figSize[0]*sf)
    plt.yticks(fontsize= domains[0].figSize[0]*sf)
    plt.grid()
    plt.tick_params(axis='both', width=figSize[0]/domains[0].defaultFigSize[0])
    plt.tick_params(axis='both', length=figSize[0]/domains[0].defaultFigSize[0]*5)
    plt.savefig(filename)
    if showPlot == True:
        plt.show()

def plotMesh(domains, mesh, filename, figSize, showPlot, meshNumbering):
    """
    plots mesh.

    Parameters
    ----------
    domains : list
        list of geometries
    mesh : object
        mesh object
    filename : str
        name of file
    figSize : list
        size of figure
    showPlot : boolean
        show plot if True
    meshNumbering : boolean
        show node and element numbers if true

    Returns
    -------
    None
    """

    plt.close("all")
    if figSize == "Default":
        figSize = domains[0].figSize

    fig, ax = plt.subplots(figsize=figSize)
    w = domains[0].width
    h = domains[0].height
    [xp,yp] = domains[0].position
    dx = w/100
    sf = 1.5
    plt.xlim(xp-w/2-dx, xp+w/2+dx)
    plt.ylim(yp-h/2-dx, yp+h/2+dx)

    plt.triplot(mesh.x, mesh.y, mesh.tri, c = 'k', lw = figSize[0]/domains[0].defaultFigSize[0])

    if (meshNumbering == "unNumbered"):
        print("Skipping Element and Node numbering")
    elif (meshNumbering == "numbered"):
        # Display element and node numbering
        fs = figSize[0]/domains[0].non * 15
        ne = len(mesh.tri[:,0])
        nn = len(mesh.x)
        for i in range(ne):
            ex = (mesh.x[mesh.tri[i,0]] + mesh.x[mesh.tri[i,1]] + mesh.x[mesh.tri[i,2]])/3
            ey = (mesh.y[mesh.tri[i,0]] + mesh.y[mesh.tri[i,1]] + mesh.y[mesh.tri[i,2]])/3
            ex = ex/(domains[0].width+2*dx)
            ey = ey/(domains[0].height+2*dx)
            ex = ex + 0.5
            ey = ey + 0.5
            ax.text(ex, ey, f"%d" % i, horizontalalignment='center', verticalalignment='center', transform=ax.transAxes, fontsize=fs, color='r')

        for i in range(nn):
            nx = mesh.x[i]
            ny = mesh.y[i]
            nx = nx/(domains[0].width+2*dx)
            ny = ny/(domains[0].height+2*dx)
            nx = nx + 0.5
            ny = ny + 0.5
            ax.text(nx, ny, f"%d" % i, horizontalalignment='center', verticalalignment='center', transform=ax.transAxes, fontsize=fs, color='b')

    plt.gca().set_aspect("equal")
    for axis in ['top', 'bottom', 'left', 'right']:
        ax.spines[axis].set_linewidth(figSize[0]/domains[0].defaultFigSize[0])
    
    plt.xlabel("x - axis [meter]", fontsize=figSize[0]/domains[0].defaultFigSize[0]*1.5*20*sf)
    plt.ylabel("y - axis [meter]", fontsize=figSize[0]/domains[0].defaultFigSize[0]*1.5*20*sf)
    plt.xticks(fontsize= figSize[0]/domains[0].defaultFigSize[0]*20*sf)
    plt.yticks(fontsize= figSize[0]/domains[0].defaultFigSize[0]*20*sf)
    plt.grid()
    plt.tick_params(axis='both', width=figSize[0]/domains[0].defaultFigSize[0])
    plt.tick_params(axis='both', length=figSize[0]/domains[0].defaultFigSize[0]*5)
    plt.savefig(filename)
    if showPlot == True:
        plt.show()

def plotContour(domains, mesh, solver, plotType, filename, colorbarPosition, figSize, showPlot):
    """
    plots solution.

    Parameters
    ----------
    domains : list
        list of geometries
    mesh : object
        mesh object
    solver : object
        the solver object
    plotType : str
        type of plot
    filename : str
        name of file
    colorbarPosition : str
        position of color bar
    figSize : list
        size of figure
    showPlot : boolean
        show plot if True

    Returns
    -------
    None
    """

    plt.close("all")
    if figSize == "Default":
        figSize = domains[0].figSize

    fig, ax = plt.subplots(figsize=figSize)
    w = domains[0].width
    h = domains[0].height
    [xp,yp] = domains[0].position
    dx = w/100
    sf = 1.5
    plt.xlim(xp-w/2-dx, xp+w/2+dx)
    plt.ylim(yp-h/2-dx, yp+h/2+dx)

    figRatio = domains[0].defaultFigSize[1]/domains[0].defaultFigSize[0]

    if (plotType == "Source"):
        triang = mtri.Triangulation(mesh.x, mesh.y, mesh.tri)
        plt.tricontourf(triang, mesh.nodeSource[:,0], 990,  cmap = "viridis")
        if solver.name == "electrostatic":
            if colorbarPosition == "top" or colorbarPosition == "bottom":
                cbar = plt.colorbar(location = colorbarPosition, fraction=0.044/figRatio, format='%.0e')
            else:
                cbar = plt.colorbar(location = colorbarPosition, fraction=0.046*figRatio, format='%.0e')
            cbar.set_label(label="Charge Density [C/m]",size=figSize[0]/domains[0].defaultFigSize[0]*30*sf)
        elif solver.name == "magnetostatic":
            if colorbarPosition == "top" or colorbarPosition == "bottom":
                cbar = plt.colorbar(location = colorbarPosition, fraction=0.044/figRatio, format='%.0e')
            else:
                cbar = plt.colorbar(location = colorbarPosition, fraction=0.046*figRatio, format='%.0e')
            cbar.set_label(label="Electric Current [Amp]",size=figSize[0]/domains[0].defaultFigSize[0]*30*sf)
        else:
            print("plotContour does not support Source type of the specified solver ", solver.name)
        
    elif (plotType == "Material"):
        triang = mtri.Triangulation(mesh.x, mesh.y, mesh.tri)
        plt.tricontourf(triang, mesh.nodeMaterial[:,0], 990,  cmap = "viridis")
        if solver.name == "electrostatic":
            if colorbarPosition == "top" or colorbarPosition == "bottom":
                cbar = plt.colorbar(location = colorbarPosition, fraction=0.044/figRatio, format='%.0e')
            else:
                cbar = plt.colorbar(location = colorbarPosition, fraction=0.046*figRatio, format='%.0e')
            cbar.set_label(label="Relative Permittivity",size=figSize[0]/domains[0].defaultFigSize[0]*30*sf)
        elif solver.name == "magnetostatic":
            if colorbarPosition == "top" or colorbarPosition == "bottom":
                cbar = plt.colorbar(location = colorbarPosition, fraction=0.044/figRatio, format='%.0e')
            else:
                cbar = plt.colorbar(location = colorbarPosition, fraction=0.046*figRatio, format='%.0e')
            cbar.set_label(label="Relative Permeability",size=figSize[0]/domains[0].defaultFigSize[0]*30*sf)
        else:
            print("plotContour does not support Material type of the specified solver ", solver.name)

    elif (plotType == "InducedSources"):
        triang = mtri.Triangulation(mesh.x, mesh.y, mesh.tri)
        ep0 = solver.physicalConstant
        if solver.name == "electrostatic":
            plt.tricontourf(triang, solver.bi[:,0]*ep0, 990,  cmap = "viridis")
            if colorbarPosition == "top" or colorbarPosition == "bottom":
                cbar = plt.colorbar(location = colorbarPosition, fraction=0.044/figRatio, format='%.0e')
            else:
                cbar = plt.colorbar(location = colorbarPosition, fraction=0.046*figRatio, format='%.0e')
            cbar.set_label(label="Charge Density [C/m]",size=figSize[0]/domains[0].defaultFigSize[0]*30*sf)
        elif solver.name == "magnetostatic":
            plt.tricontourf(triang, solver.bi[:,0]/ep0, 990,  cmap = "viridis")
            if colorbarPosition == "top" or colorbarPosition == "bottom":
                cbar = plt.colorbar(location = colorbarPosition, fraction=0.044/figRatio, format='%.0e')
            else:
                cbar = plt.colorbar(location = colorbarPosition, fraction=0.046*figRatio, format='%.0e')
            cbar.set_label(label="Electric Current [Amp]",size=figSize[0]/domains[0].defaultFigSize[0]*30*sf)
        else:
            print("plotContour does not support Source type of the specified solver ", solver.name)

    elif (plotType == "SourceDSS"):
        triang = mtri.Triangulation(mesh.x, mesh.y, mesh.tri)
        plt.tricontourf(triang, mesh.nodeSourceDss[:,0]-mesh.nodeSource[:,0], 990,  cmap = "viridis")
        if solver.name == "electrostatic":
            if colorbarPosition == "top" or colorbarPosition == "bottom":
                cbar = plt.colorbar(location = colorbarPosition, fraction=0.044/figRatio, format='%.0e')
            else:
                cbar = plt.colorbar(location = colorbarPosition, fraction=0.046*figRatio, format='%.0e')
            cbar.set_label(label="Charge Density [C/m]",size=figSize[0]/domains[0].defaultFigSize[0]*30*sf)
        elif solver.name == "magnetostatic":
            if colorbarPosition == "top" or colorbarPosition == "bottom":
                cbar = plt.colorbar(location = colorbarPosition, fraction=0.044/figRatio, format='%.0e')
            else:
                cbar = plt.colorbar(location = colorbarPosition, fraction=0.046*figRatio, format='%.0e')
            cbar.set_label(label="Electric Current [Amp]",size=figSize[0]/domains[0].defaultFigSize[0]*30*sf)
        else:
            print("plotContour does not support Source type of the specified solver ", solver.name)

    plt.gca().set_aspect("equal")
    for axis in ['top', 'bottom', 'left', 'right']:
        ax.spines[axis].set_linewidth(figSize[0]/domains[0].defaultFigSize[0])
    
    plt.xlabel("x - axis [meter]", fontsize=figSize[0]/domains[0].defaultFigSize[0]*1.5*20*sf)
    plt.ylabel("y - axis [meter]", fontsize=figSize[0]/domains[0].defaultFigSize[0]*1.5*20*sf)
    plt.xticks(fontsize= figSize[0]/domains[0].defaultFigSize[0]*20*sf)
    plt.yticks(fontsize= figSize[0]/domains[0].defaultFigSize[0]*20*sf)
    plt.grid()
    plt.tick_params(axis='both', width=figSize[0]/domains[0].defaultFigSize[0])
    plt.tick_params(axis='both', length=figSize[0]/domains[0].defaultFigSize[0]*5)
    cbar.ax.tick_params(labelsize=figSize[0]/domains[0].defaultFigSize[0]*20*sf)
    plt.savefig(filename)
    if showPlot == True:
        plt.show()

def plotContourField(domains, filename, mesh, solver, solutionName, figSize, colorbarPosition, showPlot):
    """
    plots solution.

    Parameters
    ----------
    domains : list
        list of geometries
    filename : str
        name of file
    mesh : object
        mesh object
    solver : object
        the solver object
    solutionName : str
        name of solution
    figSize : list
        size of figure
    colorbarPosition : str
        position of color bar
    showPlot : boolean
        show plot if True

    Returns
    -------
    None
    """

    solution = solver.u
    x = mesh.rx
    y = mesh.ry
    plt.close("all")
    if figSize == "Default":
        figSize = domains[0].figSize

    fig, ax = plt.subplots(figsize=figSize)
    w = domains[0].width
    h = domains[0].height
    [xp,yp] = domains[0].position
    dx = w/100
    sf = 1.5
    plt.xlim(xp-w/2-dx, xp+w/2+dx)
    plt.ylim(yp-h/2-dx, yp+h/2+dx)
    figRatio = domains[0].defaultFigSize[1]/domains[0].defaultFigSize[0]

    triang = mtri.Triangulation(mesh.x, mesh.y, mesh.tri)
    plt.tricontourf(triang, solution[:,0]-solution[:,0][0], 990,  cmap =  "viridis")

    if colorbarPosition == "top" or colorbarPosition == "bottom":
        cbar = plt.colorbar(location = colorbarPosition, fraction=0.044/figRatio, format='%.0e')
    else:
        cbar = plt.colorbar(location = colorbarPosition, fraction=0.046*figRatio, format='%.0e')

    if solver.name == "electrostatic":
        cbar.set_label(label="Potential [Volt]",size=figSize[0]/domains[0].defaultFigSize[0]*30*sf)

    elif solver.name == "magnetostatic":
        cbar.set_label(label="Potential [Tesla meter]",size=figSize[0]/domains[0].defaultFigSize[0]*30*sf)

    plt.quiver(x, y, solver.Ex, solver.Ey)
    plt.gca().set_aspect("equal")
    for axis in ['top', 'bottom', 'left', 'right']:
        ax.spines[axis].set_linewidth(figSize[0]/domains[0].defaultFigSize[0])
    
    plt.xlabel("x - axis [meter]", fontsize=figSize[0]/domains[0].defaultFigSize[0]*1.5*20*sf)
    plt.ylabel("y - axis [meter]", fontsize=figSize[0]/domains[0].defaultFigSize[0]*1.5*20*sf)
    plt.xticks(fontsize= figSize[0]/domains[0].defaultFigSize[0]*20*sf)
    plt.yticks(fontsize= figSize[0]/domains[0].defaultFigSize[0]*20*sf)
    plt.tick_params(axis='both', width=figSize[0]/domains[0].defaultFigSize[0])
    plt.tick_params(axis='both', length=figSize[0]/domains[0].defaultFigSize[0]*5)
    cbar.ax.tick_params(labelsize=figSize[0]/domains[0].defaultFigSize[0]*20*sf)
    plt.savefig(filename)
    if showPlot == True:
        plt.show()

def plotSurf(domains, filename, mesh, solver, solution, solutionName, figSize, showPlot):
    """
    plots solution.

    Parameters
    ----------
    domains : list
        list of geometries
    filename : str
        name of file
    mesh : object
        mesh object
    solver : object
        the solver object
    solution : list
        the solution
    solutionName : str
        name of solution
    figSize : list
        size of figure
    showPlot : boolean
        show plot if True

    Returns
    -------
    None
    """

    x = mesh.rx
    y = mesh.ry
    plt.close("all")
    if figSize == "Default":
        figSize = domains[0].figSize

    sf = 1.5
    fig = plt.figure(figsize=figSize)
    ax = plt.axes(projection='3d')
    ax.set_box_aspect((np.ptp(x), np.ptp(y), np.ptp(y)))
    ts = ax.plot_trisurf(x, y, np.transpose(solution)[0]-np.transpose(solution)[0][0], cmap = "viridis", edgecolor = 'grey', linewidth=0, antialiased=False)

    ax.set_xlabel("x - axis [meter]",size=figSize[0]/domains[0].defaultFigSize[0]*30*sf, labelpad=30*figSize[0]/domains[0].defaultFigSize[0])
    ax.set_ylabel("y - axis [meter]",size=figSize[0]/domains[0].defaultFigSize[0]*30*sf, labelpad=30*figSize[0]/domains[0].defaultFigSize[0])
    ax.zaxis.set_label_position("top")
    if solver.name == "electrostatic":
        ax.set_zlabel("Potential [Volt]",size=figSize[0]/domains[0].defaultFigSize[0]*30*sf, labelpad=30*figSize[0]/domains[0].defaultFigSize[0]) #, format='%.0e'

    elif solver.name == "magnetostatic":
        ax.set_zlabel("Potential [Tesla meter]",size=figSize[0]/domains[0].defaultFigSize[0]*30*sf, labelpad=30*figSize[0]/domains[0].defaultFigSize[0])

    ax.view_init(30, -110)
    fig.subplots_adjust(left=0, right=1, bottom=0, top=1)
    plt.grid()
    ax.zaxis.set_major_formatter(FormatStrFormatter('%.0e'))
    ax.xaxis.set_tick_params(labelsize=figSize[0]/domains[0].defaultFigSize[0]*20*sf)
    ax.yaxis.set_tick_params(labelsize=figSize[0]/domains[0].defaultFigSize[0]*20*sf)
    ax.zaxis.set_tick_params(labelsize=figSize[0]/domains[0].defaultFigSize[0]*20*sf)
    plt.savefig(filename)
    if showPlot == True:
        plt.show()

def plotDifference(domains, filename, mesh, solver, appxsolution, solution, figSize, showPlot):
    """
    plots difference solution.

    Parameters
    ----------
    domains : list
        list of geometries
    filename : str
        name of file
    mesh : object
        mesh object
    solver : object
        the solver object
    appxsolution : list
        the approximate solution
    solution : list
        the solution
    solutionName : str
        name of solution
    figSize : list
        size of figure
    showPlot : boolean
        show plot if True

    Returns
    -------
    None
    """

    x = mesh.rx
    y = mesh.ry
    plt.close("all")
    if figSize == "Default":
        figSize = domains[0].figSize
    sf = 1.5
    fig = plt.figure(figsize=figSize)
    ax = plt.axes(projection='3d')
    ax.set_box_aspect((np.ptp(x), np.ptp(y), np.ptp(y)))
    difference = (np.transpose(appxsolution)[0]-np.transpose(appxsolution)[0][0])-(np.transpose(solution)[0]-np.transpose(solution)[0][0])
    ts = ax.plot_trisurf(x, y, difference, cmap = "viridis", edgecolor = 'grey', linewidth=0, antialiased=False)

    ax.set_xlabel("x - axis [meter]",size=figSize[0]/domains[0].defaultFigSize[0]*30*sf, labelpad=30*figSize[0]/domains[0].defaultFigSize[0])
    ax.set_ylabel("y - axis [meter]",size=figSize[0]/domains[0].defaultFigSize[0]*30*sf, labelpad=30*figSize[0]/domains[0].defaultFigSize[0])

    if solver.name == "electrostatic":
        ax.set_zlabel("Potential [Volt]",size=figSize[0]/domains[0].defaultFigSize[0]*30*sf, labelpad=30*figSize[0]/domains[0].defaultFigSize[0])

    elif solver.name == "magnetostatic":
        ax.set_zlabel("Potential [Tesla meter]",size=figSize[0]/domains[0].defaultFigSize[0]*30*sf, labelpad=30*figSize[0]/domains[0].defaultFigSize[0])

    ax.view_init(30, -110)
    fig.subplots_adjust(left=0, right=1, bottom=0, top=1)
    plt.grid()
    ax.zaxis.set_major_formatter(FormatStrFormatter('%.0e'))
    ax.xaxis.set_tick_params(labelsize=figSize[0]/domains[0].defaultFigSize[0]*20*sf)
    ax.yaxis.set_tick_params(labelsize=figSize[0]/domains[0].defaultFigSize[0]*20*sf)
    ax.zaxis.set_tick_params(labelsize=figSize[0]/domains[0].defaultFigSize[0]*20*sf)
    plt.savefig(filename)
    if showPlot == True:
        plt.show()
